# -*- Mode: Python; coding: utf-8 -*-
# vi:si:et:sw=4:sts=4:ts=4

##
## Stoqdrivers
## Copyright (C) 2005-2007 Async Open Source <http://www.async.com.br>
## All rights reserved
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
## USA.
##
## Author(s):   Cleber Rodrigues      <cleber@globalred.com.br>
##              Henrique Romano       <henrique@async.com.br>
##              Johan Dahlin          <jdahlin@async.com.br>
##              Ronaldo Maia          <romaia@async.com.br>
##
"""
Bematech MP25 driver
"""
import re
import binascii
from decimal import Decimal
from docutils.nodes import description
from kiwi.datatypes import currency
from stoqdrivers.logs.log import traced as log
from kiwi.python import Settable
from stoqdrivers.enum import TaxType, UnitType
from stoqdrivers.exceptions import DriverError, OutofPaperError, PrinterError, \
    CommandError, CouponOpenError, HardwareFailure, PrinterOfflineError, CodigoInvalido, \
    PaymentAdditionError, ItemAdditionError, CancelItemError, CouponTotalizeError, \
    CouponNotOpenError, AlmostOutofPaper, BlockZ, PendingReduceZ, PrinterHeadUp, AuthenticationFailure
from stoqdrivers.interfaces import ICouponPrinter
from stoqdrivers.printers.base import BaseDriverConstants
from stoqdrivers.printers.capabilities import Capability
from stoqdrivers.serialbase import SerialBase
from stoqdrivers.translation import stoqdrivers_gettext
from zope.interface import implements
import datetime
import struct



_ = stoqdrivers_gettext


CASH_IN_TYPE = "SU"
CASH_OUT_TYPE = "SA"

# The comment after the commands are comparing the MP25 command with MP20.

CMD_COUPON_OPEN = 0 # MP25 aceita 2 parametros opcionais a mais
CMD_CLOSE_TILL = 5
CMD_REDUCE_Z = 5
CMD_READ_X = 6
CMD_READ_MEMORY = 8
CMD_COUPON_CANCEL = 14 # MP25 aceita 3 parametros opicionais
CMD_SUMMER_TIME = 18 # Horário de Verão
CMD_STATUS = 19
CMD_OPEN_TILL = 22 # Abrir Gaveta
CMD_ADD_VOUCHER = 25
CMD_READ_TAXCODES = 26
CMD_READ_TOTALIZERS_PARCIAL = 27 # tot. parciais nao completos
CMD_READ_TOTALIZERS = 87 # totalizadores parciais completos
CMD_READ_LAST_REDUCE = 88
CMD_GET_COUPON_SUBTOTAL = 29
CMD_GET_COUPON_NUMBER = 30
CMD_CANCEL_LAS_ITEM = 13 # cancela o ultimo item
CMD_CANCEL_ITEM = 31 # cancela um item especifico
CMD_COUPON_TOTALIZE = 32 # MP 20 Nao permite desconto simultaneo
CMD_COUPON_CLOSE = 34
CMD_READ_REGISTER = 35 # Registradores diferentes
CMD_ADD_ITEM = 63
CMD_PAYMENT_RECEIPT_OPEN = 66 # MP20 suporta somente 3 primeiro parametros
CMD_PAYMENT_RECEIPT_PRINT = 67
CMD_PAYMENT_RECEIPT_CLOSE = 21 # Parece o mesmo, mas com nomes diferentes na mp20 e 25
CMD_GERENCIAL_REPORT_OPEN = 83
CMD_GERENCIAL_REPORT_PRINT = 67
CMD_GERENCIAL_REPORT_CLOSE = 21
CMD_DISCOUNT_SURCHARGE_AFTER = 93 # Desconto/Acresciomo para o item no final do cupom

CMD_PROGRAM_PAYMENT_METHOD = 71
CMD_ADD_PAYMENT = 72

# Be carefull with these commands, They are only available for MP25 (not mp20)
CMD_CANCEL_LAST = 81
CMD_PAYMENT_RECEIPT_PRINT_DUPLICATE = 91

NAK = 21
ACK = 6
STX = 2

RETRIES_BEFORE_TIMEOUT = 10

TWOPLACES = Decimal(10) ** -2
TREPLACES = Decimal(10) ** -3

# Page 51
def isbitset(value, bit):
    # BCD crap
    return (int(value, 16) >> bit) & 1 == 1

class MP25Registers(object):
    TOTAL = 3
    TOTAL_CANCELATIONS = 4
    TOTAL_DISCOUNT = 5
    TOTAL_TESTE = 11
    COO = 6
    COO_INICIAL_FINAL = 72
    GNF = 7
    CFC = 8
    NUMBER_REDUCTIONS_Z = 9
    CRO = 10
    LAST_ITEM_ID = 12
    CLICHE = 64
    NUMBER_TILL = 14
    READ_LOJA = 15
    FISCAL_FLAGS = 17
    EMISSION_DATE = 23
    DATE_LAST_RZ = 26
    MOV_DATE = 27
    TRUNC_FLAG = 28
    TOTALIZERS = 29
    PAYMENT_METHODS = 32
    SERIAL = 40
    FIRMWARE = 41
    READ_IE = 43
    CDC = 52
    GRG = 54
    CCF = 55
    DATE_MOV_LAST_RZ = 67
    GT = 73

    # (size, bcd)
    formats = {
        TOTAL: ('9s', True),
        TOTAL_CANCELATIONS: ('7s', True),
        TOTAL_DISCOUNT: ('7s', True),
        TOTAL_TESTE: ('7s', True),
        COO: ('3s', True),
        COO_INICIAL_FINAL: ('6s', True),
        READ_IE: ('20s', False),
        READ_LOJA: ('2s', True),
        GNF: ('3s', True),
        CFC: ('2s', True),
        CDC: ('2s', True),
        GRG: ('3s', True),
        NUMBER_REDUCTIONS_Z: ('2s', True),
        CRO: ('2s', True),
        LAST_ITEM_ID: ('2s', True),
        NUMBER_TILL: ('2s', True),
        FISCAL_FLAGS: ('1s', False),
        EMISSION_DATE: ('6s', True),
        DATE_LAST_RZ: ('6s', True),
        MOV_DATE: ('3s', True),
        TRUNC_FLAG: ('1s', False),
        TOTALIZERS: ('2s', False),
        CLICHE: ('6s', False),
        #  1 + (52 * 16) + (52 * 10) + (52 * 10) + (52 * 1)
        #  1 + 832 + 520 + 520 + 52: 1925
        PAYMENT_METHODS: ('b832s520s520s52s', False),
        SERIAL: ('20s', False),
        FIRMWARE: ('3s', True),
        CCF: ('3s', True),
        DATE_MOV_LAST_RZ: ('3s', True),
        GT: ('18s', True),
        }


class MP25Constants(BaseDriverConstants):
    _constants = {
        UnitType.WEIGHT:      'Kg',
        UnitType.METERS:      'm ',
        UnitType.LITERS:      'Lt',
        UnitType.EMPTY:       '  ',
        }


class MP25Status(object):
    st1_codes = {
        128: (OutofPaperError(_("Impressora sem papel"))),
#        64: (AlmostOutofPaper(_("Impressora com pouco papel"))),
        32: (PrinterError(_("Printer clock error"))),
        16: (PrinterError(_("Printer in error state"))),
        8: (CommandError(_("First data value in CMD is not ESC (1BH)"))),
        4: (CommandError(_("Nonexistent command"))),
        2: (CouponOpenError(_("Impressora tem um cupom aberto atualmente"))),
        1: (CommandError(_("Invalid number of parameters")))}

    st2_codes = {
        128: (CommandError(_("Invalid CMD parameter"))),
        64: (HardwareFailure(_("Fiscal memory is full"))),
        32: (HardwareFailure(_("Error in CMOS memory"))),
        16: (PrinterError("Alíquota não programada")),
        8: (DriverError(_("No available tax slot"))),
        4: (CancelItemError(_("The item wasn't added in the coupon or can't "
                              "be cancelled"))),

        # 2: (PrinterError(_("Owner data (CGC/IE) not programmed on the printer"))),
        # FIXME: This shouldn't be commented. But it will break the tests.
        # Need to update the tests for all bematech printers
        1: (CommandError("Comando nao executado"))
        }

    st3_codes = {
         #0: Comando ok
         #1 Comando inválido
         #2 Erro desconhecido
         #3 Número de parâmetro inválido
         #4 Tipo de parâmetro inválido
         #5 Todas alíquotas já programadas
         #6 Totalizador não fiscal já programado
         #7 Cupom fiscal aberto
         #8 Cupom fiscal fechado
         #9 ECF ocupado
         #0 Impressora em erro
         11: (DriverError("Impressora sem papel")),
         12: (PrinterHeadUp("Impressora com cabeça levantada")),
         13: (PrinterOfflineError("Impressora OFF LINE")),
         14: (DriverError("Alíquota não programada")),
         15: (DriverError("Terminador de string faltando")),
         16: (DriverError("Acréscimo ou desconto maior que o total do cupom fiscal")),
         17: (DriverError("Cupom fiscal sem item vendido")),
         18: (DriverError("Comando nao efetivado")),
         19: (DriverError("Sem espaço para novas formas de pagamento")),
         20: (PaymentAdditionError("Forma de pagamento não programada")),
         21: (PaymentAdditionError("Índice maior que número de forma de pagamento")),
         22: (DriverError("Formas de pagamento encerradas")),
         23: (DriverError("Cupom não totalizado")),
         24: (DriverError("Comando maior que 7F")),
         25: (DriverError("Cupom fiscal aberto e sem item")),
         26: (DriverError("Cancelamento não imediatamento após")),
         27: (DriverError("Cancelamento já efetuado")),
         28: (DriverError("Comprovante de crédito ou débito não permitido ou já emitido")),
         29: (DriverError("Meio de pagamento não permite TEF")),
         30: (DriverError("Sem comprovante não fiscal aberto")),
         31: (DriverError("Comprovante de crédito ou débito já aberto")),
         32: (DriverError("Reimpressão não permitida")),
         33: (DriverError("Comprovante não fiscal já aberto")),
         34: (DriverError("Totalizador não fiscal não programado")),
         35: (DriverError("Cupom não fiscal sem item vendido")),
         36: (DriverError("Acréscimo e desconto maior que total CNF")),
         37: (DriverError("Meio de pagamento não indicado")),
         38: (DriverError("Meio de pagamento diferente do total de recebimento")),
         39: (DriverError("Não permitido mais de uma sangria ou suprimento")),
         40: (DriverError("Relatório já programado")),
         41: (DriverError("Relatório gerencial não programado")),
         42: (DriverError("Relatório gerencial não permitido")),
         43: (CouponNotOpenError("MFD não inicializada")),
         44: (DriverError("MFD ausente")),
         45: (DriverError("MFD sem número de série")),
         46: (DriverError("MFD já inicializada")),
         47: (DriverError("MFD lotada")),
         48: (DriverError("Cupom não fiscal aberto")),
         49: (DriverError("Memória fiscal desconectada")),
         50: (DriverError("Memória fiscal sem número de série da MFD")),
         51: (DriverError("Memória fiscal lotada")),
         52: (DriverError("Data inicial inválida")),
         53: (DriverError("Data final inválida")),
         54: (DriverError("Contador de Redução Z inicial inválido")),
         55: (DriverError("Contador de Redução Z final inválido")),
         56: (DriverError("Erro de alocação")),
         57: (DriverError("Dados do RTC incorretos")),
         58: (DriverError("Data anterior ao último documento emitido")),
         59: (DriverError("Fora de intervenção técnica")),
         60: (DriverError("Em intervenção técnica")),
         61: (DriverError("Erro na memória de trabalho")),
         #62: (DriverError("Já houve movimento no dia")),
         63: (BlockZ("Bloqueio por RZ")),
         64: (DriverError("Forma de pagamento aberta")),
         65: (DriverError("Aguardando primeiro proprietário")),
         66: (DriverError("Aguardando RZ")),
         67: (DriverError("ECF ou loja igual a zero")),
         68: (DriverError("Cupom adicional não permitido")),
         69: (DriverError("Desconto maior que total vendido em ICMS")),
         70: (DriverError("Recebimento não fiscal nulo não permitido")),
         71: (DriverError("Acréscimo ou desconto maior que total não fiscal")),
         72: (DriverError("Memória fiscal lotada para novo cartucho")),
         73: (DriverError("Erro de gravação na MF")),
         74: (DriverError("Erro de gravação na MFD")),
         75: (DriverError("Dados do RTC anteriores ao último documento armazenado")),
         76: (DriverError("Memória fiscal sem espaço para gravar leituras da MFD")),
         77: (DriverError("Memória fiscal sem espaço para gravar versão do SB")),
         78: (DriverError("Descrição igual a default não permitido")),
         79: (DriverError("Extrapolado número de repetições permitidas")),
         80: (DriverError("Segunda via do comprovante de crédito ou débito não permitido")),
         81: (DriverError("Parcelamento fora da sequência")),
         82: (DriverError("Comprovante de crédito ou débito aberto")),
         83: (DriverError("Texto com sequência de ESC inválida")),
         84: (DriverError("Texto com sequência de ESC incompleta")),
         85: (DriverError("Venda com valor nulo")),
         86: (DriverError("Estorno de valor nulo")),
         87: (DriverError("Forma de pagamento diferente do total da sangria")),
         88: (DriverError("Redução não permitida em intervenção técnica")),
         89: (DriverError("Aguardando RZ para entrada em intervenção técnica")),
         90: (DriverError("Forma de pagamento com valor nulo não permitido")),
         91: (ItemAdditionError("Acréscimo ou desconto maior que valor do ítem")),
         92: (DriverError("Autenticação não permitida")),
         93: (DriverError("Timeout na validação")),
         94: (DriverError("Comando não executado em impressora bilhete de passagem")),
         95: (DriverError("Comando não executado em impressora de Cupom Fiscal")),
         96: (DriverError("Cupom Não Fiscal fechado")),
         97: (DriverError("Parâmetro não ASCII em campo ASCII")),
         98: (DriverError("Parâmetro não ASCII num em campo ASCII num")),
         99: (DriverError("Tipo de transporte inválido")),
         100: (DriverError("Data e hora inválida")),
         101: (CodigoInvalido("Sem relatório gerencial")),
         102: (DriverError("Número do totalizador não fiscal inválido")),
         103: (DriverError("Parâmetro de acréscimo ou desconto inválido")),
         104: (DriverError("Acréscimo ou desconto em sangria ou suprimento não permitido")),
         105: (DriverError("Número do relatório gerencial inválido")),
         106: (DriverError("Forma de pagamento fonte não programada")),
         107: (DriverError("Forma de pagamento destino não programada")),
         108: (DriverError("Estorno maior que forma pagamento")),
         109: (DriverError("Caracter numérico na codificação GT não permitido")),
         110: (DriverError("Erro na inicialização da MF")),
         111: (DriverError("Nome do totalizador em branco não permitido")),
         112: (DriverError("Data e hora anteriores ao último doc armazenado")),
         113: (DriverError("Parâmetro de acréscimo ou desconto inválido")),
         114: (DriverError("Item anterior aos trezentos últimos")),
         115: (CancelItemError("Item não existe ou já cancelado")),
         116: (DriverError("Código com espaços não permitido")),
         117: (DriverError("Descrição sem caracter alfabético não permitido")),
         118: (DriverError("Acréscimo maior que valor do item")),
         119: (DriverError("Desconto maior que valor do item")),
         120: (DriverError("Desconto em ISS não permitido")),
         121: (DriverError("Acréscimo em item já efetuado")),
         122: (DriverError("Desconto em item já efetuado")),
         123: (DriverError("Erro na Memória Fiscal chamar credenciado")),
         124: (DriverError("Aguardando gravação na Memória Fiscal")),
         125: (DriverError("Caracter repetido na codificação do GT")),
         126: (DriverError("Versão já gravada na memória fiscal")),
         127: (DriverError("Estouro de capacidade no cheque (não utilizado neste modelo)")),
         128: (DriverError("Timeout na leitura do cheque (não utilizado neste modelo)")),
         129: (CouponOpenError("Mês inválido (não utilizado neste modelo)")),
         130: (DriverError("Coordenada inválida (não utilizado neste modelo)")),
         131: (DriverError("Sobreposição de texto (não utilizado neste modelo)")),
         132: (DriverError("Sobreposição de texto no valor (não utilizado neste modelo)")),
         133: (DriverError("Sobreposição de texto no extenso (não utilizado neste modelo)")),
         134: (DriverError("Sobreposição de texto no favorecido (não utilizado neste modelo)")),
         135: (DriverError("Sobreposição de texto na localidade (não utilizado neste modelo)")),
         136: (DriverError("Sobreposição de texto no opcional (não utilizado neste modelo)")),
         137: (DriverError("Sobreposição de texto no dia (não utilizado neste modelo)")),
         138: (DriverError("Sobreposição de texto no mês (não utilizado neste modelo)")),
         139: (DriverError("Sobreposição de texto no ano (não utilizado neste modelo)")),
         140: (DriverError("Usando MFD de outro ECF")),
         141: (DriverError("Primeiro dado diferente de esc ou 1C")),
         142: (DriverError("Não permitido alterar sem intervenção técnica")),
         143: (DriverError("Dados da última RZ corrompidos")),
         144: (DriverError("Comando não permitido no modo inicialização")),
         145: (DriverError("Aguardando acerto de relógio")),
         146: (DriverError("MFD já inicializada para outra memoria fiscal")),
         147: (DriverError("Aguardando acerto do relógio ou desbloqueio pelo teclado")),
         148: (DriverError("Valor forma de pagamento maior que máximo permitido")),
         149: (DriverError("Razão social em branco")),
         150: (DriverError("Nome de fantasia em branco")),
         151: (DriverError("Endereço em branco")),
         152: (DriverError("Estorno de CDC não permitido")),
         153: (DriverError("Dados do proprietário iguais ao atual")),
         154: (DriverError("Estorno de forma de pagamento não permitido")),
         155: (DriverError("Descricao forma de pagamento igual já programada")),
         156: (DriverError("Acerto de horário de verão só imediatamente após Redução Z")),
         157: (DriverError("Intervenção técnica não permitida MF reservada para Redução Z")),
         158: (DriverError("Senha CNPJ inválida")),
         159: (DriverError("Timeout na inicialização de nova MF")),
         160: (DriverError("Dados da MFD não encontrados")),
         161: (DriverError("Sangria ou Suprimento devem ser únicos no comprovante não fiscal")),
         162: (CodigoInvalido("Índice da forma de pagamento nulo não permitido")),
         163: (DriverError("UF de destino inválida")),
         164: (DriverError("Tipo de transporte incompatível com a UF de destino")),
         165: (DriverError("Descrição do primeiro item do Bilhete de passagem diferente de “TARIFA”")),
         166: (DriverError("Aguardando impressão de cheque ou autenticação (não utilizado neste modelo)")),
         167: (DriverError("Programação de CNPJ e IE preenchido com espaços em branco não permitido")),
         168: (DriverError("Programação de UF preenchido com espaços em branco não permitido")),
         169: (DriverError("Número de impressões da fita detalhe nesta intervenção técnica esgotado")),
         170: (PaymentAdditionError("Cupom fiscal já sub-totalizado")),
         171: (PaymentAdditionError("Cupom não sub-totalizado")),
         172: (PaymentAdditionError("Acréscimo em sub-total já efetuado")),
         173: (PaymentAdditionError("Desconto em sub-total já efetuado")),
         174: (PaymentAdditionError("Acréscimo com valor nulo não permitido")),
         175: (PaymentAdditionError("Desconto com valor nulo não permitido")),
         176: (PaymentAdditionError("Cancelamento de acréscimo ou desconto em sub-total não permitido")),
         177: (DriverError("Data inválida")),
         178: (DriverError("Valor do cheque nulo não permitido (não utilizado neste modelo)")),
         179: (DriverError("Valor do cheque inválido (não utilizado neste modelo)")),
         180: (DriverError("Cheque sem a informação da localidade não permitido (não utilizado neste modelo)")),
         181: (DriverError("Cancelamento de acréscimo em item não permitido")),
         182: (DriverError("Cancelamento de desconto em item não permitido")),
         183: (DriverError("Número máximo de itens atingido")),
         184: (DriverError("Número do item nulo não permitido")),
         185: (DriverError("Mais que duas alíquotas diferentes no Bilhete de Passagem não permitido")),
         186: (DriverError("Acréscimo ou Desconto em item não permitido")),
         187: (DriverError("Cancelamento de acréscimo ou desconto em item não permitido")),
         188: (DriverError("Clichê já impresso")),
         189: (DriverError("Texto opcional do cheque excedeu o máximo de linhas permitido (não utilizado neste modelo)")),
         190: (DriverError("Impressão automática no verso não disponível neste equipamento")),
         191: (DriverError("Timeout na inserção do cheque (não utilizado neste modelo)")),
         192: (DriverError("Overflow na capacidade de texto do comprovante de débito ou crédito")),
         193: (DriverError("Programação de espaços entre cupons menor que o mínimo permitido")),
         194: (DriverError("Equipamento não possui leitor de cheque (não utilizado neste modelo)")),
         195: (DriverError("Programação de alíquota com valor nulo não permitido")),
         196: (DriverError("Parâmetro baud rate inválido")),
         197: (DriverError("Configuração permitida somente pela porta do fisco")),
         198: (DriverError("Valor total do item excede 11 dígitos")),
         199: (DriverError("Programação da moeda com espaços em branco não permitido")),
         200: (DriverError("Casa decimais devem ser programadas com 2 ou 3")),
         201: (DriverError("Não permite cadastrar usuários diferentes na mesma MFD")),
         202: (DriverError("Identificação do consumidor não permitido para sangria ou suprimento")),
         203: (DriverError("Casas decimais em quantidade maior do que a permitida")),
         204: (DriverError("Casas decimais do unitário maior do que o permitido")),
         205: (DriverError("Posição reservada para ICMS")),
         206: (DriverError("Posição reservada para ISS")),
         207: (DriverError("Todas as alíquotas com a mesma vinculação não permitido")),
         208: (DriverError("Data de embarque anterior a data de emissão")),
         209: (DriverError("Alíquota para ISSQN não permitida sem inicialização da Inscrição Municipal")),
         210: (DriverError("Pacote do clichê gráfico fora da seqüência")),
         211: (DriverError("Espaço para armazenamento do clichê gráfico esgotado")),
         212: (DriverError("Clichê gráfico não disponível para confirmação")),
         213: (DriverError("CRC do clichê gráfico diferente do informado")),
         214: (DriverError("Intervalo inválido")),
         215: (DriverError("Usuário já programado")),
         217: (DriverError("Detectada abertura do equipamento"))
        }

    def __init__(self, reply):
        self.st1, self.st2, self.st3 = reply[-3:]


    @property
    def open(self):
        return bool(self.st1 == 2)
    @property
    def pagamento_existe(self):
        return bool(self.st1 == 66)
    @property
    def out_paper(self):
        return self.st1 == 64
    @property
    def block_z(self):
        return self.st3 == 66

    @property
    def off_line(self):
        return self.st3 == 13

    def _check_error_in_dict(self, error_codes, value):
        for key in error_codes:
            if key & value:
                raise error_codes[key]

    def check_error(self):
#        print "status: st1=%s st2=%s st3=%s" % (self.st1, self.st2, self.st3)
        log.debug("status: st1=%s st2=%s st3=%s" % (self.st1, self.st2, self.st3))
        if self.st1 != 0:
            self._check_error_in_dict(self.st1_codes, self.st1)

        if self.st2 != 0:
            self._check_error_in_dict(self.st2_codes, self.st2)

            # first bit means not executed, look in st3 for more
            if self.st2 & 1 and self.st3:
                if self.st3 in self.st3_codes:
                    raise self.st3_codes[self.st3]

    def check_error_two(self):
        if self.st1 not in [0, 64]:
            self.check_error()

#
# Helper functions
#

def bcd2dec(data):
    return int(''.join(['%02x' % ord(i) for i in data]))

def bcd2hex(data):
    return ''.join(['%02x' % ord(i) for i in data])

def dec2bcd(dec):
    return chr(dec % 10 + (dec / 10) * 16)

def dec2bin(n, trim= -1):
    a = ""
    while n > 0:
        if n % 2 == 0:
            a = "0" + a
        else:
            a = "1" + a
        n /= 2

    if trim != -1:
        if len(a) < trim:
            a = ("0" * (trim - len(a))) + a
    return a

#
# Driver implementation
#

class MP25(SerialBase):
    implements(ICouponPrinter)
    CMD_PROTO = 0x1c

    supported = True
    model_name = "Bematech MP25 FI"
    coupon_printer_charset = "cp850"
    supports_duplicate_receipt = True
    identify_customer_at_end = False
    registers = MP25Registers
    reply_format = '<b%sbbH'
    status_size = 3

    EOL_DELIMIT = '\n'

    def __init__(self, port, consts=None):
        self._consts = consts or MP25Constants
        port.set_options(read_timeout=3, write_timeout=5)
        SerialBase.__init__(self, port)
        # XXX: Seems that Bematech doesn't contains any variable with the
        # coupon remainder value, so I need to manage it by myself.
        self.remainder_value = Decimal("0.00")
        self._reset()

    def _reset(self):
        self._customer_name = ''
        self._customer_document = ''
        self._customer_address = ''

    def _create_packet(self, command):
        """
        Create a 'pre-package' (command + params, basically) and involves
        it around STX, NB and CS::
           1     2           n           2
        +-----+------+-----------------+----+
        | STX |  NB  | 0x1C + command  | CS |
        +-----+------+-----------------+----+

        Where:

        STX: 'Transmission Start' indicator byte
        NB: 2 bytes, big endian length of command + CS (2 bytes)
        CS: 2 bytes, big endian checksum for command
        """

        command = chr(self.CMD_PROTO) + command
        return struct.pack('<bH%dsH' % len(command),
                           STX,
                           len(command) + 2,
                           command,
                           sum([ord(i) for i in command]))

    def _read_reply(self, size):
        a = 0
        data = ''
        ack = False
        while True:
            if a > RETRIES_BEFORE_TIMEOUT:
                raise DriverError("ECF Desconectada/Desligada ou Sem Papel")

            a += 1
            if not ack:
                reply = self.read(1)
            else: # caso retornou o ack mas esta demorando no retorno do resto do comando
                reply = self.read(size-1) # tamanho menos o ACK

            if not reply:
                continue

            if reply == chr(6): # ACK = comando foi executado na impressora
                ack = True
                data += reply
                reply = self.read(size-1) # tamanho menos o ACK
            elif reply == 21 or reply == chr(15): # NAK = erro na execução do comando ou timeout na recepcao dos dados pela impressora
                raise DriverError("Erro na execução do comando/ECF Desligada ou Sem Papel")
            elif not ack:  # o elemento lido nao eh ACK e nem NAC ou seja, deve ser lixo no buffer
                a = 0
                continue

            data += reply
            if len(data) < size:
                continue

            log.debug("<<< %r (%d bytes)" % (data, len(data)))
            return data

    def _check_error(self, retval=None):
        status = self.get_status(retval)
        status.check_error()

    def _send_command(self, command, *args, **kwargs):
        log.debug("_send_command(%r, %r %r')" % (command, args, kwargs))
        fmt = ''
        if 'response' in kwargs:
            fmt = kwargs.pop('response')

        raw = False
        if 'raw' in kwargs:
            raw = kwargs.pop('raw')

        if kwargs:
            raise TypeError("Invalid kwargs: %r" % (kwargs,))

        cmd = chr(command)
        for arg in args:
            if isinstance(arg, int):
                cmd += chr(arg)
            elif isinstance(arg, str):
                cmd += arg
            else:
                raise NotImplementedError(type(arg))

        data = self._create_packet(cmd)

        self.write(data)

        format = self.reply_format % fmt
        # Condição implementada abaixo é um tratamento p/ não prosseguir c/ retorno invalido da ecf.
        # Aki é verificado se o primeiro bit é igual 06H e se o comando enviado pertencia a lista de
        # comandos.
        while True:
            reply = self._read_reply(struct.calcsize(format))
            if reply[0] != chr(6) and command in [29, 35]:
                self.write(data)
            else:
                break
        if command == 71:
            return reply
        retval = struct.unpack(format, reply)

        if raw:
            #self._check_error(retval)
            return retval

        response = retval[1:-self.status_size]
        if len(response) == 1:
            response = response[0]
        return response

    def retorno_de_variaveis(self, var, tam):
        retval = self._send_command(35, var, response='%ds' % tam)
        return binascii.b2a_hex(retval)

    def ler_ultima_reducao(self):
        retval = self._send_command(62, 55, response='308s')
        retval = binascii.b2a_hex(retval)
        return retval

    def _read_register(self, reg):
        try:
            fmt, bcd = self.registers.formats[reg]
        except KeyError:
            raise NotImplementedError(reg)

        value = self._send_command(CMD_READ_REGISTER, reg, response=fmt)
        if bcd:
            value = bcd2dec(value)
        return value

    #
    # Helper methods
    #

    def _get_coupon_subtotal(self):
        subtotal = self._send_command(CMD_GET_COUPON_SUBTOTAL, response='7s')

        if subtotal:
            return Decimal(bcd2dec(subtotal)) / Decimal("1e2")
        # Busted subtotal
        return Decimal("0.0")

    def _get_last_item_id(self):
        try:
            res = self._read_register(self.registers.LAST_ITEM_ID)
        except Exception as e:
            log.error(str(e))
            res = 999

        return res

    def _get_coupon_number(self):
        coupon_number = self._send_command(CMD_GET_COUPON_NUMBER, response='3s')
        return bcd2dec(coupon_number)

    def get_status(self, val=None, cmd_status=None):
        log.debug("get_status(%r, %r)" % (val, cmd_status))

        # if not val:
        #     val = self._send_command(1, raw=True)

        if cmd_status is None:
            cmd_status = CMD_STATUS

        if val is None:
            val = self._send_command(cmd_status, raw=True)
            print 'st', val
            #val = self._send_command(1, raw=True)
        return MP25Status(val)

    def _add_voucher(self, type, value, text=''):
        assert len(type) == 2
        status = self.get_status()
        if status.open:
            self._send_command(CMD_COUPON_CANCEL)
        ret = self._send_command(CMD_ADD_VOUCHER, type, "%014d%s" % (int(value * Decimal('1e2')), text), raw=True)
        self.get_status(ret).check_error()

    def _configure_payment_methods(self, descricao):
        # Do one at a time, if you need it.
        ret = self._send_command(CMD_PROGRAM_PAYMENT_METHOD,
                                 '%-16s' % descricao[:80], response='3', raw=False)
        try:
            int(ret[1:3])
            return ret[1:3]
        except:
            print 'Finalizadora cheia'
            return False
        #ret = self._send_command(CMD_PROGRAM_PAYMENT_METHOD,
        #                         '%-16s1' % 'Cartao Debito', raw=True)
        #ret = self._send_command(CMD_PROGRAM_PAYMENT_METHOD,
        #                         '%-16s1' % 'Cheque', raw=True)

    #
    # This implements the ICouponPrinter Interface
    #

    def summarize(self):
        """ Prints a summary of all sales of the day """
        return self._send_command(CMD_READ_X, raw=True)

    def get_totalizadores(self, total=False):
        '''
        Esta funcao busca cada totalizador da ecf e seus valores
        '''
        total_cancelations = self._read_register(self.registers.TOTAL_CANCELATIONS)
        total_discount = self._read_register(self.registers.TOTAL_DISCOUNT)

        length, names = self._send_command(CMD_READ_TAXCODES, response='b32s') # quantidade e nome dos totalizadores
        status = self._read_register(self.registers.TOTALIZERS)
        status = struct.unpack('>H', status)[0]
        if total: # pega todos totalizadores (ver manual da bematec)
            cmd_totalizadores = CMD_READ_TOTALIZERS
            response = '436s'
        else: # nao pega todos totalizadores
            cmd_totalizadores = CMD_READ_TOTALIZERS_PARCIAL
            response = '219s'

        values = self._send_command(cmd_totalizadores, response=response)
        taxes = []
        for i in range(length):
            if 1 << 15 - i & status != 0:
                type = 'ISS'
            else:
                type = 'ICMS'

            name = bcd2hex(names[i * 2:i * 2 + 2])
            value = bcd2dec(values[i * 7:i * 7 + 7])
            taxes.append(('%02d' % (i+1), value / Decimal(100), type))

        taxes.append(('ACRESC', bcd2dec(values[182:182 + 7]) / Decimal(100), 'ISS'))
        taxes.append(('ACRESC', bcd2dec(values[161:161 + 7]) / Decimal(100), 'ICMS'))
        taxes.append(('CANC', bcd2dec(values[189:189 + 7]) / Decimal(100), 'ISS'))
        taxes.append(('CANC', total_cancelations / Decimal(100), 'ICMS'))
        taxes.append(('DESC', bcd2dec(values[175:175 + 7]) / Decimal(100), 'ISS'))
        taxes.append(('DESC', total_discount / Decimal(100), 'ICMS'))
        taxes.append(('I', bcd2dec(values[112:119]) / Decimal(100), 'ICMS'))
        taxes.append(('N', bcd2dec(values[119:126]) / Decimal(100), 'ICMS'))
        taxes.append(('F', bcd2dec(values[126:133]) / Decimal(100), 'ICMS'))
        return taxes


    def close_till(self, previous_day=False):
        """ Close the till for the day, no other actions can be done after this
        is called.
        """
        # a impressao da Leitura X precisa ser feita diariamente. Ao utilizar o
        # parâmetro previous_day a Leitura X só será impressa quando o caixa
        # não é fechado no mesmo dia
        self._send_command(CMD_REDUCE_Z)

    def till_add_cash(self, value, totalizador=0, meio=0):
        self._add_voucher(CASH_IN_TYPE, value, text='')

    def till_remove_cash(self, value, cod=None, text=''):
        self._add_voucher(CASH_OUT_TYPE, value, text)

    def till_read_memory(self, start, end, full=False):
        i = 'i' if not full else 'I'
        self._send_command(CMD_READ_MEMORY,
                           '%s%s%s' % (start.strftime('%d%m%y'),
                                      end.strftime('%d%m%y'),
                                      i))

    def till_read_memory_to_serial(self, start, end, full=False):
        r = 'r' if not full else 'R'
        self._send_command(CMD_READ_MEMORY,
                           '%s%s%s' % (start.strftime('%d%m%y'),
                                       end.strftime('%d%m%y'),
                                       r)
                           )

        data = ''
        while True:
            c = self.read(1)
            if ord(c) == 0x03:
                break

            data += unicode(c, self.coupon_printer_charset)

        return data

    def till_read_memory_by_reductions_to_serial(self, start, end, full=False):
        r = 'r' if not full else 'R'
        ret = self._send_command(CMD_READ_MEMORY,
                                 '%06d%06d%s' % (start, end, r))
        data = ''
        while True:
            c = self.read(1)
            if ord(c) == 0x03:
                break

            data += unicode(c, self.coupon_printer_charset)

        return data


    def till_read_memory_by_reductions(self, start, end, full=False):
        i = 'i' if not full else 'I'
        self._send_command(CMD_READ_MEMORY,
                           '%06d%06d%s' % (start, end, i))

    def coupon_identify_customer(self, customer, address, document):
        self._customer_name = customer
        self._customer_document = document
        self._customer_address = address

    def coupon_is_customer_identified(self):
        return len(self._customer_document) > 0

    def coupon_open(self):
        """ This needs to be called before anything else """
        if self._customer_name:
            ret = self._send_command(CMD_COUPON_OPEN,
                                "%-29s%-30s%-80s" % (self._customer_document,
                                                     self._customer_name,
                                                     self._customer_address), raw=True)
        else:
            ret = self._send_command(CMD_COUPON_OPEN, raw=True)
        self.get_status(ret).check_error()
        return ret
    def coupon_cancel(self):
        """ Can only be called when a coupon is opened. It needs to be possible
        to open new coupons after this is called.
        """
        retcmd = self._send_command(CMD_COUPON_CANCEL, raw=True)
        self.get_status(retcmd).check_error()


    def cancel_last_coupon(self):
        """Cancel the last non fiscal coupon or the last sale."""
        self._send_command(CMD_CANCEL_LAST)

    def coupon_close(self, message=""):
        """  This can only be called when the coupon is open, has items added,
        payments added and totalized is called. It needs to be possible to open
        new coupons after this is called.
        """
        # FIXME This will break the drivers recorded tests. Record new ones.
        message = re.sub(r'NOTA LEGAL:', chr(27) + chr(69) + r'NOTA LEGAL:' + chr(27) + chr(70) , message)
        self._send_command(CMD_COUPON_CLOSE, message)
        self._reset()
        return True

    def coupon_add_item(self, code, description, price, taxcode,
                        quantity=Decimal("1.0"), unit=UnitType.EMPTY,
                        discount=Decimal("0.0"), markup=Decimal("0.0"),
                        unit_desc=""):
        """ The ECF must be configured to round instead to truncate.
        When truncating, the value may be lower then the one calculated
        by stoq. In this case, the payments added will be higher than the ECF
        expects, and a cents change will be printed.
        """
        if taxcode in ['F', 'I', 'N']:
            taxcode += taxcode
        if unit == UnitType.CUSTOM:
            unit = unit_desc
        else:
            unit = self._consts.get_value(unit)

        data = ("%s"     # taxcode
                "%09d"     # value
                "%07d"     # quantity
                "%010d"    # discount
                "%010d"    # markup
                "%022d"    # padding
                "%2s"      # unit
                "%-48s\0"  # code
                "%-200s\0" # description
                % (taxcode.rjust(2, '0'),
                   price * Decimal("1e3"),
                   quantity * Decimal("1e3"),
                   discount * Decimal("1e2"),
                   markup * Decimal("1e2"),
                   0, unit, code, description))
        di = datetime.datetime.now()

        retval = self._send_command(CMD_ADD_ITEM, data, raw=True)

        self.get_status(retval).check_error()
        return self._get_last_item_id()

    def desc_acres_item(self, desc_acres, item, valor):
        '''
        Efetua desconto acrescimo no item
        :param desc_acres: 'A' para acrescimo, 'D' para desconto, 1 byte
        :param item: numero de sequencia do item no cupom, 3 bytes
        :param valor: valor do desconto/acrescimo,  8 bytes
        :return:
        '''

        parametros = desc_acres + \
                     "%03d" % item + \
                     "%08d" % (Decimal(valor).quantize(TWOPLACES) * Decimal("1e2"))
        return self._send_command(CMD_DISCOUNT_SURCHARGE_AFTER, parametros)

    def cancel_last_item(self):
        return self._send_command(CMD_CANCEL_LAS_ITEM)

    def coupon_cancel_item(self, item_id=None):
        """ Cancel an item added to coupon; if no item id is specified,
        cancel the last item added. """
        last_item = self._get_last_item_id()
        if item_id is None:
            return self.cancel_last_item()

        elif item_id not in xrange(1, last_item + 2):
            raise CancelItemError("There is no such item with ID %r"
                                  % item_id)
        retcmd = self._send_command(CMD_CANCEL_ITEM, "%04d" % (item_id,), raw=True)
        self.get_status(retcmd).check_error()

    def coupon_add_payment(self, payment_method, value, description=u""):
        #payment_method = '01'
        retval = self._send_command(CMD_ADD_PAYMENT, "%s%014d%s" % (
        payment_method, int(value * Decimal('1e2')), description[:80]), raw=True)
        if retval[3] != 23:
            self.get_status(retval).check_error()
        self.remainder_value -= value
        if self.remainder_value < 0.0:
            self.remainder_value = Decimal("0.0")
        return self.remainder_value

    def coupon_totalize(self, discount=currency(0), markup=currency(0), taxcode=TaxType.NONE):

        if discount:
            type = 'd'
            value = discount
        elif markup:
            type = 'a'
            value = markup
        else:
            # Just to use the StartClosingCoupon in case of no discount/markup
            # be specified.
            type = 'a'
            value = 0
        vlr = '%s%014d' % (type, int(value * Decimal('1e2')))
        retval = self._send_command(CMD_COUPON_TOTALIZE, vlr, raw=True)
        if retval[2] == 1:
            self.get_status(retval).check_error()

        totalized_value = self._get_coupon_subtotal()
        self.remainder_value = totalized_value
        return totalized_value

    def get_payment_receipt_identifier(self, method):
        # We don't need one.
        return None

    def _get_payment_description(self, method_id):
        constants = self.get_payment_constants()
        for value, name in constants:
            if value == method_id:
                return name

    def gerencial_report_open(self, id=None):
        if id:
            st = self._send_command(CMD_GERENCIAL_REPORT_OPEN, '%02d' % id, raw=True)
        else:
            st = self._send_command(CMD_GERENCIAL_REPORT_PRINT, raw=True)
        self.get_status(st).check_error()

    def gerencial_report_print(self, text):
        imprime_linha = ''
        for line in text.split('\n'):
            if len(imprime_linha + '\n' + line) > 620: # limite de caracteres do comando CMD_GERENCIAL_REPORT_PRINT
                self._send_command(CMD_GERENCIAL_REPORT_PRINT, imprime_linha)
                imprime_linha = ''

            imprime_linha += '\n' + line

        self._send_command(CMD_GERENCIAL_REPORT_PRINT, imprime_linha) # imprime a o restante do texto

    def gerencial_report_close(self):
        self._send_command(CMD_GERENCIAL_REPORT_CLOSE)

    def payment_receipt_open(self, identifier, coo, method, value):
        method = self._get_payment_description(method)
        if not method:
            raise DriverError('Looks like this payment method '
                              'is not configured in the printer')

        value = int(value * 100)
        self._send_command(CMD_PAYMENT_RECEIPT_OPEN,
                           '%-16s%014d%06d' % (method, value, coo))

    def get_corta_papel(self): # as MPs 4000 e 2100 nao possuem comando para cortar papel
        return False

    def payment_receipt_print(self, text):
        imprime_linha = ''
        for line in text.split('\n'):
            if len(imprime_linha + '\n' + line) > 620: # limite de caracteres do comando CMD_PAYMENT_RECEIPT_PRINT
                self._send_command(CMD_PAYMENT_RECEIPT_PRINT, imprime_linha)
                imprime_linha = ''

            if line == 'corta_papel':
                if imprime_linha:
                    self._send_command(CMD_PAYMENT_RECEIPT_PRINT, imprime_linha)
                    imprime_linha = ''

                if not self.get_corta_papel():
                    imprime_linha = '\n------------------------------------------------\n\n'

                continue

            imprime_linha += '\n' + line

        self._send_command(CMD_PAYMENT_RECEIPT_PRINT, imprime_linha) # imprime a o restante do texto

    def payment_receipt_close(self):
        self._send_command(CMD_PAYMENT_RECEIPT_CLOSE)

    def payment_receipt_print_duplicate(self):
        self._send_command(CMD_PAYMENT_RECEIPT_PRINT_DUPLICATE)

    def get_capabilities(self):
        return dict(
            item_code=Capability(max_len=14),
            item_id=Capability(digits=4),
            items_quantity=Capability(min_size=1, digits=4, decimals=3),
            item_price=Capability(digits=6, decimals=2),
            item_description=Capability(max_len=200),
            payment_value=Capability(digits=12, decimals=2),
            promotional_message=Capability(max_len=320),
            payment_description=Capability(max_len=48),
            customer_name=Capability(max_len=30),
            customer_id=Capability(max_len=28),
            customer_address=Capability(max_len=80),
            add_cash_value=Capability(min_size=0.1, digits=12, decimals=2),
            remove_cash_value=Capability(min_size=0.1, digits=12, decimals=2),
            )

    def get_constants(self):
        return self._consts

    def query_status(self):
        #return '\x02\x05\x00\x1b#(f\x00'
        query = self._create_packet(chr(CMD_READ_REGISTER) +
                                    chr(self.registers.SERIAL))
        return query

    def status_reply_complete(self, reply):
        return len(reply) == 25

    def get_serial(self):
        return self._read_register(self.registers.SERIAL).strip('\x00')

    def get_ccf(self):
        return self._read_register(self.registers.CCF)

    def get_coo(self):
        return self._read_register(self.registers.COO)

    def get_coo_inicial_final(self):
        return self._read_register(self.registers.COO_INICIAL_FINAL)

    def get_read_ie(self):
        return self._read_register(self.registers.READ_IE)

    def get_num_loja(self):
        return self._read_register(self.registers.READ_LOJA)

    def get_num_ecf(self):
        return self._read_register(self.registers.NUMBER_TILL)

    def get_gnf(self):
        return self._read_register(self.registers.GNF)

    def get_cfc(self):
        return self._read_register(self.registers.CFC)
    def get_grg(self):
        return self._read_register(self.registers.GRG)
    def get_cdc(self):
        return self._read_register(self.registers.CDC)

    def get_crz(self):
        return self._read_register(self.registers.NUMBER_REDUCTIONS_Z)

    def get_tax_constants(self):
        status = self._read_register(self.registers.TOTALIZERS)
        status = struct.unpack('>H', status)[0]

        length, data = self._send_command(CMD_READ_TAXCODES, response='b32s')

        service = False
        constants = []
        for i in range(16):
            value = bcd2dec(data[i * 2:i * 2 + 2])
            if not value:
                continue

            if 1 << 15 - i & status == 0:
                tax = TaxType.CUSTOM
            else:
                tax = TaxType.SERVICE
            constants.append((tax,
                              '%02d' % (i + 1,),
                              Decimal(value) / 100))

        constants.extend([
            (TaxType.SUBSTITUTION, 'FF', None),
            (TaxType.EXEMPTION, 'II', None),
            (TaxType.NONE, 'NN', None),
            ])

        return constants

    def get_payment_constants(self):
        status = self._read_register(self.registers.PAYMENT_METHODS)[1]
        methods = []
        for i in range(20):
            method = status[i * 16:i * 16 + 16]
            if method != '\x00' * 16:
                methods.append(('%02d' % (i + 1), method.strip()))
        return methods

    def get_grande_total(self):
        grande_total = self._read_register(self.registers.TOTAL)
        grande_total = grande_total / Decimal(100)
        return grande_total

    def get_date_last_mov_rz(self):
        '''
        :return: Retorna a data do movimento da ultima reducao Z emitida
        '''
        str_date_mov_last_rz = str(self._read_register(self.registers.DATE_MOV_LAST_RZ))
        try:
            if len(str_date_mov_last_rz) == 5:
                str_date_mov_last_rz = '0' + str_date_mov_last_rz

            date_last_mov_rz = datetime.datetime(year=2000 + int(str_date_mov_last_rz[4:6]),
                                                 month=int(str_date_mov_last_rz[2:4]),
                                                 day=int(str_date_mov_last_rz[:2])).date()
        except Exception as e:
            # quando emite um Z que nao eh pendente em um dia que nao teve movimentacao, a data do movimento
            # do dele nao existe.
            log.error('get_date_last_mov_rz: %s' % str(e))
            date_last_mov_rz = 0

        return date_last_mov_rz


    def get_sintegra(self):
        opening_date = self._read_register(self.registers.EMISSION_DATE)
        cro = self._read_register(self.registers.CRO)
        # FIXME: This is being fetched before the actual reduction, so the value will be wrong by
        # -1
        crz = self._read_register(self.registers.NUMBER_REDUCTIONS_Z)
        coo = self._get_coupon_number()
        ccf = self.get_ccf()
        gnf = self.get_gnf()
        grg = self.get_grg()
        cdc = self.get_cdc()
        serial = self.get_serial()
        total_cancelations = self._read_register(self.registers.TOTAL_CANCELATIONS)
        total_discount = self._read_register(self.registers.TOTAL_DISCOUNT)

        registers = self._send_command(62, 55, response='308s')
        if self.get_status().st3 == 143:
            raise self.get_status().st3_codes[143]

        #  cupom inicial
        try:
            coupon_start = int(bcd2hex(registers)[568:568 + 6]) + 1
        except:
            coupon_start = bcd2hex(registers)[568:568 + 6]

        # cupom final
        coupon_end = coo

        grande_total = self._read_register(self.registers.TOTAL)
        grande_total = grande_total / Decimal(100)

        # este tratamento eh devido ao erro que da ao tentar ler o GT do ultimo Z, a string dos
        # registradores retornada de "self._send_command(62, 55, response='308s')" vem toda
        # desconfigurada, neste percebi que pode ser quando a ECF eh nova e nao houve Z ainda
        try:
            total_bruto = bcd2dec(registers[1:10]) / Decimal(100)
        except:
            try:
                registers = self._send_command(62, 55, response='308s')
                total_bruto = bcd2dec(registers[1:10]) / Decimal(100)
            except Exception as e:
                log.error('GT ultimo Z: %s', str(e))
                total_bruto = Decimal(0)

        length, names = self._send_command(CMD_READ_TAXCODES, response='b32s')
        status = self._read_register(self.registers.TOTALIZERS)
        status = struct.unpack('>H', status)[0]

        values = self._send_command(CMD_READ_TOTALIZERS, response='436s')

        date = '%012d' % opening_date
        taxes = []
        for i in range(length):

            if 1 << 15 - i & status != 0:
                type = 'ISS'
            else:
                type = 'ICMS'

            name = bcd2hex(names[i * 2:i * 2 + 2])
            value = bcd2dec(values[i * 7:i * 7 + 7])
            taxes.append((name, value / Decimal(100), type))

        taxes.append(('ACRESC', bcd2dec(values[182:182 + 7]) / Decimal(100), 'ISS'))
        taxes.append(('ACRESC', bcd2dec(values[161:161 + 7]) / Decimal(100), 'ICMS'))
        taxes.append(('CANC', bcd2dec(values[189:189 + 7]) / Decimal(100), 'ISS'))
        taxes.append(('CANC', total_cancelations / Decimal(100), 'ICMS'))
        taxes.append(('DESC', bcd2dec(values[175:175 + 7]) / Decimal(100), 'ISS'))
        taxes.append(('DESC', total_discount / Decimal(100), 'ICMS'))
        taxes.append(('I', bcd2dec(values[112:119]) / Decimal(100), 'ICMS'))
        taxes.append(('N', bcd2dec(values[119:126]) / Decimal(100), 'ICMS'))
        taxes.append(('F', bcd2dec(values[126:133]) / Decimal(100), 'ICMS'))

        serial_id = '%03d' % self._read_register(self.registers.NUMBER_TILL)

        str_date_last_rz = str(self._read_register(self.registers.DATE_LAST_RZ))
        try:
            if len(str_date_last_rz) == 11:
                str_date_last_rz = '0' + str_date_last_rz

            date_last_rz = datetime.date(year=2000 + int(str_date_last_rz[4:6]),
                                         month=int(str_date_last_rz[2:4]),
                                         day=int(str_date_last_rz[:2]))
        except Exception as e:
            date_last_rz = {0: str_date_last_rz, 1: str(e)}

        date_last_mov_rz = self.get_date_last_mov_rz()
        if not date_last_mov_rz:
            date_last_mov_rz = date_last_rz


        return Settable(
            date_last_rz=date_last_rz,
            date_last_mov_rz=date_last_mov_rz,
            opening_date=date_last_rz,
            serial=serial,
            serial_id=serial_id,
            coupon_start=coupon_start,
            coupon_end=coupon_end,
            cro=cro,
            crz=crz,
            coo=coo,
            ccf=ccf,
            gnf=gnf,
            grg=grg,
            cdc=cdc,
            period_total=grande_total - total_bruto,
            total=grande_total,
            taxes=taxes
        )

    def get_tax(self):
        total_cancelations = self._read_register(self.registers.TOTAL_CANCELATIONS)
        total_discount = self._read_register(self.registers.TOTAL_DISCOUNT)
        # Avbr function TACBrECFBematech.GetVendaBruta

        length, names = self._send_command(CMD_READ_TAXCODES, response='b32s')
        status = self._read_register(self.registers.TOTALIZERS)
        status = struct.unpack('>H', status)[0]

        values = self._send_command(CMD_READ_TOTALIZERS, response='436s')

        #        values2 = self._send_command(27, response='218s')
        #        return (bcd2hex(values), bcd2hex(values2))

        taxes = []
        for i in range(length):

            if 1 << 15 - i & status != 0:
                type = 'ISS'
            else:
                type = 'ICMS'

            name = bcd2hex(names[i * 2:i * 2 + 2])
            value = bcd2dec(values[i * 7:i * 7 + 7])
            taxes.append((name, value / Decimal(100), type))

        taxes.append(('ACRESC', bcd2dec(values[182:182 + 7]) / Decimal(100), 'ISS'))
        taxes.append(('ACRESC', bcd2dec(values[161:161 + 7]) / Decimal(100), 'ICMS'))
        taxes.append(('CANC', bcd2dec(values[189:189 + 7]) / Decimal(100), 'ISS'))
        taxes.append(('CANC', total_cancelations / Decimal(100), 'ICMS'))
        taxes.append(('DESC', bcd2dec(values[175:175 + 7]) / Decimal(100), 'ISS'))
        taxes.append(('DESC', total_discount / Decimal(100), 'ICMS'))
        taxes.append(('I', bcd2dec(values[112:119]) / Decimal(100), 'ICMS'))
        taxes.append(('N', bcd2dec(values[119:126]) / Decimal(100), 'ICMS'))
        taxes.append(('F', bcd2dec(values[126:133]) / Decimal(100), 'ICMS'))

        return taxes


    def get_firmware_version(self):
        """Return the firmware version"""
        # REGISTER IS AN INTEGER: 10000 and shoud be formated as 01.00.00
        ret = self._read_register(self.registers.FIRMWARE)
        ret = '%0*d' % (6, ret)
        firmware = "%s:%s:%s" % (ret[0:2], ret[2:4], ret[4:6])
        return firmware
    ##
    # By Christian Douglas
    ##
    def set_summertime(self, value=True):
        ret_cmd = self._send_command(CMD_SUMMER_TIME, raw=True)
        self._check_error(ret_cmd)
        return True

    def get_summertime(self):
        ret_cmd = bcd2hex(self._read_register(self.registers.FISCAL_FLAGS))
        return isbitset(str(ret_cmd), 2)


    def open_cash_drawer(self, sec=0):
        return self._send_command(CMD_OPEN_TILL)

    def get_status_cash_drawer(self):
        retval = self._send_command(23, response='1s', raw=True)
        result = True if retval[1] == '\xff' else False
        return result

    def get_date_sb(self):
        ret_cmd = self._read_register(self.registers.CLICHE)
        print ret_cmd, binascii.b2a_hex(ret_cmd)

    def get_data_mov(self):
        log.debug("get_data_mov")
        ret_cmd = self._read_register(self.registers.MOV_DATE)
        if ret_cmd == 0:
            ret_cmd = 10100
        return datetime.datetime.strptime('%06d' % ret_cmd, "%d%m%y").date()

    def get_ecf_date(self):
        log.debug("get_ecf_date")
        ret_cmd = self._read_register(self.registers.EMISSION_DATE)
        if ret_cmd == 0:
            ret_cmd = 10100000000
        return datetime.datetime.strptime('%012d' % ret_cmd, "%d%m%y%H%M%S")

    def get_last_reduce(self):
        rethex = binascii.b2a_hex(self._send_command(88, response='621s'))
        GT_INI_FIM = self._read_register(self.registers.GT) or ''
        GT_INI = float(str(GT_INI_FIM)[:18] or 0)
        GTANT = Decimal(str( GT_INI / 100.00))
        retval = {'CRO': int(rethex[2:6]),
                  'CRZ': int(rethex[6:10]),
                  'COO': int(rethex[10:16]),
                  'GNF': int(rethex[16:22]),
                  'CCF': int(rethex[22:28]),
                  'GRG': int(rethex[28:34]),
                  'CFD': int(rethex[34:40]),
                  'NFC': int(rethex[40:44]),
                  'CFC': int(rethex[44:48]),
                  'COM': rethex[48:168],
                  'CER': rethex[166:288],
                  'CDC': int(rethex[288:292]),
                  'NCN': int(rethex[292:296]),
                  'CCDC': int(rethex[296:300]),
                  'GT': Decimal(rethex[300:318]) / Decimal(100),
                  'VBD': Decimal(rethex[300:318]) / Decimal(100) - GTANT}
        return retval

    def get_pagamento_existe(self):
        ret_cmd = bcd2hex(self._read_register(self.registers.FISCAL_FLAGS))
        return isbitset(str(ret_cmd), 1)
