# -*- Mode: Python; coding: iso-8859-1 -*-
# vi:si:et:sw=4:sts=4:ts=4
from kiwi.python import Settable
import re
##
## Stoqdrivers
## Copyright (C) 2005 Async Open Source <http://www.async.com.br>
## All rights reserved
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
## USA.
##
## Author(s):   Henrique Romano <henrique@async.com.br>
##
"""
Daruma FS2100 driver
"""

import operator
from decimal import Decimal
import time
import datetime
from stoqdrivers.printers.daruma.FS345 import FS345, CMD_GET_TAX_CODES
from stoqdrivers.enum import UnitType, TaxType
from stoqdrivers.exceptions import DriverError, CommandError, CloseCouponError
import random
from stoqdrivers.logs.log import traced as log

ENQ = 05
ACK = 06
LF  = 10
CR  = 13
ESC = 27
FS  = 28
GS  = 29
FF  = 255

CMD_ACIONA_GUILHOTINA = 123
CMD_ADD_ITEM = 201
CMD_DISCOUNT_SURCHARGE = 202 # desconto/acrescimo no item depois que o mesmo foi vendido
CMD_READ_REGISTER = 200
CMD_GET_TAX_CODES = 231
CMD_GET_DATES = 250
# REGISTRADORES
REG_GT = 01
REG_CRO = 23
REG_CRZ = 24
REG_SERIAL = 78

TWOPLACES = Decimal(10) ** -2
TREPLACES = Decimal(10) ** -3

class FS2100(FS345):
    model_name = "Daruma FS 2100"

    def coupon_add_item(self, code, description, price, taxcode,
                        quantity=Decimal("1.0"), unit=UnitType.EMPTY,
                        discount=Decimal("0.0"),
                        surcharge=Decimal("0.0"), unit_desc=""):

        TAX_DICT = {'F': 17,
                    'I': 19,
                    'N': 21
        }

        if taxcode in TAX_DICT.keys():
            taxcode = TAX_DICT.get(taxcode)
        else:
            taxcode = taxcode.rjust(2, '0') 

        if surcharge:
            d = 3
            E = surcharge
        else:
            d = 1
            E = discount
#        E = int(E)
        price = price
        quantity = quantity

        # The minimum size of the description, when working with one line for
        # description; if 0, write in multiple lines, if necessary.
        desc_size = 0

        if unit != UnitType.CUSTOM:
            unit = self._consts.get_value(unit)
        else:
            unit = "%2s" % unit_desc[:2] # units must be 2 byte size strings

        # XXX: We need test correctly if the price's calcule is right (we
        # don't can do it right now since the manual isn't so clean).
        data = ('%02s'  # Tributary situation
                '%07d' # Quantity
                '%08d' # Unitary price
                '%d'   # 0=Discount(%) 1=Discount($) 2=Surcharge(%) 3=Surcharge($)
                '%011d' # Discount/Surcharge value
#                '%07d' # *Padding* (since we have discount/surcharge only in %)
                '%02d' # Description size
                '%14s' # Code
                '%3s'  # Unit of measure
                '%s'   # Product descriptio?!?
                '\xff' # EOF
                % (taxcode, int(quantity * Decimal('1e3')),
                   int(price * Decimal('1e2')), d,
                   int(E * Decimal('1e2')),
                   desc_size, code, unit, description[:233]))
        print '>', data
        value = self.send_new_command('F', CMD_ADD_ITEM, data)
        return int(value[8:11])

    def desc_acres_item(self, desc_acres, item, valor):
        if desc_acres == 'D':
            desc_acres = 1  # 0 = % desc; 1 = R$ desc; 2 = % acres; 3 = R$ acres  (1 bytes)
        else:
            desc_acres = 3

        item = "%03d" % item # (3 bytes)
        valor = "%011d" % (Decimal(valor).quantize(TWOPLACES) * Decimal("1e2")) # Valor em centavos (11 bytes)
        data = "%s%s%s" % (item, desc_acres, valor)

        self.send_new_command('F', CMD_DISCOUNT_SURCHARGE, data)
        return True

    def send_command(self, command, extra=''):
        """As seen on ACBr code

        Nas Darumas com MFD, em algumas situa��es o ECF pode ficar
        temporariamente inoperantente, enquanto a compacta��o da MFD est�
        sendo efetuada. Nessa situa��o, o ECF retorna os seguintes c�digos
        de erro: 35 - Relogio Inoperante ou 99 (n�o documentado).

        Segundo o Suporte T�cnico da Daruma, quando este problema ocorre,
        devemos aguardar at� que ele consiga responder corretamente.

        Esta rotina ir� tentar enviar o comando por 10 vezes. Caso ela
        recebe os erros 35 ou 99... ele aguarda 100 milisegundos e tenta um
        novo envio...
        """
        if command == 1:
            return False

        for t in range(10):
            try:
                retval = FS345.send_command(self, command, extra)
            except DriverError, e:
                if e.code == 99 or e.code == 35:
                    log.debug('FS2100 >>> Error 99. Sleeping for 0.1')
                    time.sleep(0.1)
                    continue

                raise
            break

        # After we have tried to send the command T times, we need to read
        # all the responses. This is really crap
        for i in range(t):
            reply = self.readline()
            log.debug('Ignoring reply: %r' % reply)

        return retval

    def _check_response(self, retcode, raw):
        # Compatible with the fs345/fs2100

        compatible_error = retcode[3:6]
        extended_error = retcode[7:9]
        warning_code = retcode[6:8]
        log.debug('FS2100 >>> Error %s - Extended: %s - Warning: %s' %
                    (compatible_error, extended_error, warning_code))
        if int(compatible_error):
            # Mimic FS345 error format
            error_code = int(compatible_error)
            self.handle_error(error_code, raw)

    def send_new_command(self, prefix, command, extra='', ignore_error=False):
        """ This method is used to send especific commands to model FS2100.
        Note that the main differences are the prefix (0x1c + 'F', since we
        will use a function of the FS2100 superset) and the checksum, that
        must be included in the end of all the functions of this superset.
        """
        data = chr(command) + extra

        log.debug('FS2100 >>> %r %d' % (data, len(data)))

        data = chr(FS) + prefix + data

        checksum = reduce(operator.xor, [ord(d) for d in data], 0)
        self.write(data + chr(checksum))

        retval = self.readline()
        # After the CR, there is still one byte for the checksum
        retval_checksum = self.read(1)

        if not ignore_error:
            self._check_response(retval, data)
        return retval[1:]

    def _get_compatibility_mode(self):
        return self.send_new_command('R', 200, '138')

    def get_tax(self):
        sintegra = self.get_sintegra()
        return sintegra.taxes

    def get_tax_constants(self):
        tax_codes = self.send_command(CMD_GET_TAX_CODES)[1:]
        lower_case = 'abcdefghijklmnop'

        constants = []
        for i in range(14):
            reg = tax_codes[i*5]
            if reg in 'ABCDEFGHIJKLMNOP':
                tax_type = TaxType.CUSTOM
            elif reg in lower_case:
                tax_type = TaxType.SERVICE
            else:
                raise AssertionError(reg)
            value = tax_codes[i*5+1:i*5+5]
            if value == '////':
                continue
            constants.append((tax_type,
                              #'T' + reg.lower(),
                              '%02d' % (lower_case.index(reg.lower()) + 1),
                              Decimal(value.replace('.', '')) / 100))

        # These definitions can be found on Page 60
        constants.extend([
            (TaxType.SUBSTITUTION,   '17', None), # F1
            (TaxType.EXEMPTION,      '19', None), # I1
            (TaxType.NONE,           '21', None), # N1
            ])

        return constants

    # POR CHRISTIAN DOUGLAS
    def _read_registers(self, reg, ignore_error=True):
        reg = '%.3d' % reg
        retval = self.send_new_command('R', 200, reg, ignore_error)
        return retval[4:]

    def get_corta_papel(self):
        self._read_registers(CMD_ACIONA_GUILHOTINA)

    def get_serial(self):
        retval = self._read_registers(78)
        return retval[:-1]

    def get_coo(self):
        retval = self._read_registers(26)
        return int(retval)

    def get_ccf(self):
        retval = self._read_registers(30)
        return int(retval)

    def get_crz(self):
        retval = self._read_registers(24)
        return int(retval)
    def get_grg(self):
        retval = self._read_registers(33)
        return int(retval)
    def get_cdc(self):
        retval = self._read_registers(45)
        return int(retval)
    def get_cro(self):
        retval = self._read_registers(23)
        return int(retval)

    def get_gnf(self):
        retval = self._read_registers(28)
        return int(retval)

    def get_data_mov(self):
        retval = self._read_registers(70)
        dia = int(retval[:2])
        mes = int(retval[2:4])
        ano = int(retval[4:8])
        return datetime.datetime(ano, mes, dia).date()


    def get_ecf_date(self):
        retval = self._read_registers(66)
        dia = int(retval[:2])
        mes = int(retval[2:4])
        ano = int(retval[4:8])
        hora = int(retval[8:10])
        minutos = int(retval[10:12])
        segundos = int(retval[12:])
        return datetime.datetime(ano, mes, dia, hora, minutos, segundos)

    def get_date_sb(self):
        retval = self._read_registers(85)
        return datetime.datetime.strptime(retval, "%d%m%Y%H%M%S")

    def _get_coupon_subtotal(self):
        retval = self._read_registers(47)
        retval = Decimal('%s.%s' % (retval[:-2], retval[10:]))
        return retval


    def _send_command(self, command, raw=''):
        return self.send_command(command, extra=raw)

    def _add_payment(self, payment_method, value, description=''):
        rv = self.send_new_command('F', 209, '%.02d%012d%s\xff' % (int(payment_method),
                                                                   int(value * Decimal('1e2')),
                                                                   description[:48]), ignore_error=False)
        # FIXME: Why and when does this happen?
        #        Avoids/Fixes bug 3467 at least
        return rv #Decimal(rv) / Decimal('1e2')

    def _configure_payment_methods(self, descricao):
        ''' Metodo para verificar as finalizadoras cadastradas na ecf e rotornar o index de acordo com a descrição
        informada.'''
        retval = self._read_registers(126)
        retval = retval.replace('\xff', '')
        i = 15
        x = 0
        ind = 0
        finalizadoraNaEcf = {}
        while True:
            ret = retval[x:i] # Descri��o da finalizadora na ecf
            if not ret:
                break
            x = i
            i += 15
            ind += 1 # index na ecf
            finalizadoraNaEcf[ret] = str(ind).rjust(2, '0')
        try:
            result = finalizadoraNaEcf[descricao.ljust(15, ' ')]
        except:
            result =  False
        return result


    def coupon_add_payment(self, payment_method, value, description=''):
        
        self._verify_coupon_open()
        return self._add_payment(payment_method, value, description)

    def get_summertime(self):
        retval = self._read_registers(108)
        return bool(int(retval))

    def open_cash_drawer(self, sec):
        return self.send_new_command('N', 206, '%.03d' % sec)
    
    def get_status_cash_drawer(self):
        s = self._get_status()
        ret = self.status_check(s, 1, 3)
        return ret 
        
    def summarize(self):
        self.send_new_command('F', 235, '%d' % 0)

    def coupon_close(self, message=''):
        self._check_status()
        self._verify_coupon_open()

        if (self._customer_name or self._customer_address or self._customer_document):
            customer_name = self._customer_name or _("No client")
            customer_document = self._customer_document or _("No document")
            customer_address = self._customer_address or _("No address")
            self.send_command(201,
                              "%- 84s%- 84s%- 84s" % (customer_name,
                                                      customer_address,
                                                      customer_document))
        try:
            message = re.sub(r'NOTA LEGAL:', chr(27) + 'G1' + r'NOTA LEGAL:' + chr(27) + 'G0' , message)
            self.send_new_command('F', 210, '0%s\xff' % message, ignore_error=True)
        except DriverError as e:
            raise CloseCouponError(("Nao foi possivel encerrar o cupom"))
        self._reset()
        return True

    def till_remove_cash(self, value, cod=None, text=''):
        data = "%011d%s\xff" % (int(value * Decimal('1e2')), text)
        self.send_new_command('F', 227, data)

    def till_add_cash(self, value, totalizador=0, meio=0):
        data = "%011d\xff" % int(value * Decimal('1e2'))
        self.send_new_command('F', 236, data)

    def close_till(self, previous_day=False):
        date = time.strftime('%d%m%y%H%M%S', time.localtime())
        self.send_new_command('F', 234, date)

    def set_summertime(self, value):
        raw = '%d' % int(value)
        retval = self.send_new_command('C', 200, raw)
        return retval

    def gerencial_report_open(self, id=0):
        return self.send_new_command('F', 230, '%.02d' % id, ignore_error=True)

    def gerencial_report_print(self, text):
        for line in text.split('\n'):
            self.send_new_command('F', 231, line + chr(255), ignore_error=True)

    def gerencial_report_close(self):
        self.send_new_command('F', 232, ignore_error=True)

    def get_sintegra(self):
        registries = self._get_registers()
        fiscal_registries = self._get_fiscal_registers()

        tax_codes = self.send_command(CMD_GET_TAX_CODES)[1:]

        taxes = []
        for i in range(14):
            reg = tax_codes[i*5+1:i*5+5]
            if reg == '////':
                continue
            reg = reg.replace('.', '')

            if tax_codes[i*5] in 'ABCDEFGHIJKLMNOP':
                type = 'ICMS'
            else:
                type = 'ISS'

            sold = fiscal_registries[88+(i*14):102+(i*14)]
            taxes.append((reg, Decimal(sold)/100, type))

        taxes.append(('ACRESC', Decimal(fiscal_registries[313:326]) / 100, 'ICMS'))
        taxes.append(('DESC', Decimal(fiscal_registries[19:32]) / 100, 'ICMS'))
        taxes.append(('CANC', Decimal(fiscal_registries[33:46]) / 100, 'ICMS'))
        taxes.append(('I', Decimal(fiscal_registries[47:60]) / 100, 'ICMS'))
        taxes.append(('N', Decimal(fiscal_registries[61:74]) / 100, 'ICMS'))
        taxes.append(('F', Decimal(fiscal_registries[75:88]) / 100, 'ICMS'))

        total_sold = sum(value for _, value, _ in taxes)

        old_total = Decimal(fiscal_registries[:18]) / 100
        cancelled = Decimal(fiscal_registries[33:46]) / 100
        period_total = total_sold

        opening_date = self.get_data_mov()

        identifier = self.get_serial()
        return Settable(
             opening_date=opening_date,
             date_last_rz=opening_date,
             date_last_mov_rz=opening_date,
             serial=identifier[1:9],
             serial_id=int(identifier[13:17]),
             coupon_start=int(registries[:6]),
             coupon_end=int(registries[7:12]),
             cro=self.get_cro(),
             crz=self.get_crz(),
             coo=self.get_coo(),
             ccf=self.get_ccf(),
             gnf=self.get_gnf(),
             grg=self.get_grg(),
             cdc=self.get_cdc(),
             period_total=period_total,
             total=period_total + old_total,
             taxes=taxes)


    def get_status(self, val=None, cmd_status=None):
        status = self._get_status()
        class Status(object):
            def __init__(self, parent=None):
                self.status = status
                self.parent = parent
            @property
            def open(self):
                return self.parent.status_check(status, 4, 2)
            @property
            def pagamento_existe(self):
                return self.parent.status_check(status, 1, 2)
            @property
            def out_paper(self):
                return self.parent.status_check(status, 2, 0)
            @property
            def block_z(self):
                return self.parent.status_check(status, 2, 1)

            @property
            def off_line(self):
                st = self.parent.status_check(status, 2, 3)
                return st
            def check_error(self):
                self.parent._check_status()
            def check_error_two(self):
                if self.open:
                    self.parent._check_status()

        return Status(self)

    def payment_receipt_open(self, identifier, coo, method, value):
        print 'new', identifier, method, coo, value
        value = int(value * 100)
        data = '%s%s%06d%012d\xff\xff\xff' % (method, '01', coo, value)
        self.send_new_command('F', 212, data)
        
    def payment_receipt_print(self, text):
        for line in text.split('\n'):
            self.send_new_command('F', 213, line + '\xff')


    def payment_receipt_close(self):
        self.send_new_command('F', 214)

    def get_date_last_mov_rz(self):
        '''
        :return: Retorna a data do movimento da ultima reducao Z emitida
        '''
        ultimo_z = self.get_last_reduce()
        data = ultimo_z.get('date_last_rz')
        try:
            if len(data) == 5:
                data = '0' + data

            data = datetime.datetime(year=int(data[4:8]),
                                     month=int(data[2:4]),
                                     day=int(data[:2])).date()
        except Exception as e:
            log.error('get_date_last_mov_rz: %s' % str(e))
            return 0

        return data

    def get_last_reduce(self):
        '''
        Pega os dados da ultima reducao Z emiditda
        :return:
        '''
        retcmd = self._read_registers(140)
        print retcmd
        gtatual = self.get_grande_total()
        retval = {
            'date_last_rz': str(retcmd[0:8]),
            'GT': Decimal(retcmd[8:26]) / Decimal('100'),
            'VBD': Decimal(retcmd[8:26]) / Decimal('100') - Decimal(retcmd[26:44]) / Decimal('100'),
            'CRO': int(retcmd[922:926]),
            'CRZ': int(retcmd[926:930]),
            'CRZR': int(retcmd[930:934]),
            'COO': int(retcmd[934:940]),
            'GNF': int(retcmd[940:946]),
            'CCF': int(retcmd[946:952]),
            'CVC': int(retcmd[952:958]),
            'GRG': int(retcmd[958:964]),
            'CFD': int(retcmd[964:970]),
            'CBP': int(retcmd[970:976]),
            'NFC': int(retcmd[980:984]),
            'CMV': int(retcmd[988:992]),
            'CFC': int(retcmd[992:996]),
            'CNC': int(retcmd[996:1000]),
            'CBC': int(retcmd[1000:1004]),
            'NCN': int(retcmd[1004:1008]),
            'CDC': int(retcmd[1008:1012]),
            'CON': int(retcmd[1012:1092]),
            'CER': int(retcmd[1092:1172])
            #'NFC': retcmd[40:44],
            #'CFC': retcmd[44:48],
            #'COM': retcmd[48:168],
            #'CER': retcmd[166:288],
            #'CDC': retcmd[288:292],
            #'NCN': retcmd[292:296],
            #'CCDC': retcmd[296:300],
        }
        return retval

    def get_read_ie(self):
        retval = self._read_registers(91)
        return retval.rstrip()

    def get_num_loja(self):
        retval = self._read_registers(129)
        return retval

    def get_num_ecf(self):
        retval = self._read_registers(107)
        return retval