# -*- Mode: Python; coding: utf-8 -*-
# vi:si:et:sw=4:sts=4:ts=4

#
# Stoqdrivers
# Copyright (C) 2005-2007 Async Open Source <http://www.async.com.br>
# All rights reserved
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
# USA.
#
# Author(s):   Johan Dahlin     <jdahlin@async.com.br>
#              Henrique Romano  <henrique@async.com.br>
#

import datetime
from decimal import Decimal
from kiwi.datatypes import currency
from kiwi.argcheck import number
from stoqdrivers.logs.log import traced as log
from stoqdrivers.exceptions import (CloseCouponError, PaymentAdditionError,
                                    AlreadyTotalized, InvalidValue)
from stoqdrivers.enum import PaymentMethodType, TaxType, UnitType
from stoqdrivers.printers.base import BasePrinter
from stoqdrivers.printers.capabilities import capcheck
from stoqdrivers.utils import encode_text
from stoqdrivers.translation import stoqdrivers_gettext

_ = stoqdrivers_gettext

#
# Extra data types to argcheck
#
TWOPLACES = Decimal(10) ** -2


class taxcode(number):
    @classmethod
    def value_check(cls, name, value):
        if value not in (TaxType.NONE, TaxType.ICMS, TaxType.SUBSTITUTION,
                         TaxType.EXEMPTION):
            raise ValueError("%s must be one of TaxType.* constants" % name)


class unit(number):
    @classmethod
    def value_check(cls, name, value):
        if value not in (UnitType.WEIGHT, UnitType.METERS, UnitType.LITERS,
                         UnitType.EMPTY, UnitType.CUSTOM):
            raise ValueError("%s must be one of UNIT_* constants" % name)


class payment_method(number):
    @classmethod
    def value_check(cls, name, value):
        if value not in (PaymentMethodType.MONEY, PaymentMethodType.CHECK,
                         PaymentMethodType.CUSTOM):
            raise ValueError("%s must be one of *_PM constants" % name)


#
# FiscalPrinter interface
#

class FiscalPrinter(BasePrinter):
    def __init__(self, brand=None, model=None, device=None, config_file=None,
                 *args, **kwargs):

        BasePrinter.__init__(self, brand, model, device, config_file, *args,
                             **kwargs)
        self._has_been_totalized = False
        self.payments_total_value = Decimal("0.0")
        self.totalized_value = Decimal("0.0")
        self._capabilities = self._driver.get_capabilities()
        self._charset = self._driver.coupon_printer_charset
        self._serial = None

    def refresh(self):
        print dir(self)
        # self.refresh()

    def get_capabilities(self):
        return self._capabilities

    def _format_text(self, text):
        return encode_text(text, self._charset)

    @capcheck(basestring, basestring, basestring)
    def identify_customer(self, customer_name, customer_address, customer_id):
        log.info('> CMD: identify_customer(customer_name=%r, customer_address=%r, customer_id=%r)'
                 % (customer_name, customer_address, customer_id))

        res = self._driver.coupon_identify_customer(self._format_text(customer_name),
                                                    self._format_text(customer_address),
                                                    self._format_text(customer_id))
        log.info('< RET: %s' % str(res))
        return res

    def coupon_is_customer_identified(self):
        log.info('> CMD: coupon_is_customer_identified()')
        res = self._driver.coupon_is_customer_identified()
        log.info('< RET: %s' % str(res))
        return res

    def open(self):
        log.info('> CMD: coupon_open()')
        res = self._driver.coupon_open()
        log.info('< RET: ' + str(res))
        return res

    @capcheck(basestring, basestring, Decimal, str, Decimal, unit, Decimal, Decimal, basestring)
    def add_item(self, item_code, item_description, item_price, taxcode,
                 items_quantity=Decimal("1.0"), unit=UnitType.EMPTY, discount=Decimal("0.0"),
                 surcharge=Decimal("0.0"), unit_desc=""):

        log.info("> CMD: add_item(code=%r, description=%r, price=%r, taxcode=%r, quantity=%r, "
                 "unit=%r, discount=%r, surcharge=%r, unit_desc=%r)" %
                 (item_code, item_description, item_price, taxcode, items_quantity, unit, discount,
                  surcharge, unit_desc))

        if self._has_been_totalized:
            err = "O Cupom já foi totalizado, você não pode mais adicionar itens"
            log.error('< ERR: ' + err)
            raise AlreadyTotalized(err)

        if discount and surcharge:
            err = "discount and surcharge can not be used together"
            log.error('< ERR: ' + err)
            raise TypeError(err)

        elif unit != UnitType.CUSTOM and unit_desc:
            err = "You can't specify the unit description if you aren't using UnitType.CUSTOM constant."
            log.error('< ERR: ' + err)
            raise ValueError(err)

        elif unit == UnitType.CUSTOM and not unit_desc:
            err = "You must specify the unit description when using UnitType.CUSTOM constant."
            log.error('< ERR: ' + err)
            raise ValueError(err)

        elif unit == UnitType.CUSTOM and len(unit_desc) != 2:
            err = "unit description must be 2-byte sized string"
            log.error('< ERR: ' + err)
            raise ValueError(err)

        if not item_price:
            err = "The item value must be greater than zero"
            log.error('< ERR: ' + err)
            raise InvalidValue(err)

        retval = self._driver.coupon_add_item(self._format_text(item_code),
                                              self._format_text(item_description),
                                              item_price, taxcode, items_quantity,
                                              unit, discount, surcharge,
                                              unit_desc=self._format_text(unit_desc))
        log.info('< RET: ' + str(retval))
        return retval

    def desc_acres_item(self, desc_acres, item, valor):
        log.info('> CMD: desc_acres_item(desc_acres=%s, item=%s, valor=%s)' % (desc_acres, item, valor))
        res = self._driver.desc_acres_item(desc_acres, item, valor)
        log.info('< RET: %s' % str(res))
        return res

    @capcheck(Decimal, Decimal, taxcode)
    def totalize(self, discount=currency(0), surcharge=currency(0),
                 taxcode=TaxType.NONE):
        log.info('> CMD: totalize(discount=%r, surcharge=%r, taxcode=%r)' % (
            discount, surcharge, taxcode))

        if discount and surcharge:
            err = "discount and surcharge can not be used together"
            log.error('< ERR: ' + err)
            raise TypeError(err)

        if discount > self.get_coupon_subtotal():
            err = "Valor do Desconto/Acrescimo nao pode ser maior que o Total do Cupom"
            log.error('< ERR: ' + err)
            raise ValueError(err)

        if surcharge and taxcode == TaxType.NONE:
            err = "to specify a surcharge you need specify its tax code"
            log.error('< ERR: ' + err)
            raise ValueError(err)

        res = self._driver.coupon_totalize(discount, surcharge, taxcode)
        self._has_been_totalized = True
        self.totalized_value = res

        log.info("< RET: %s\n%s" % (str(res), "FIM DO TOTALIZE ================================================================"))

        return res

    def configure_payment_methods(self, descricao=u''):
        log.info('> CMD: configure_payment_methods(descricao=%r)' % descricao)
        res = self._driver._configure_payment_methods(descricao)
        log.info('< RET: %s' % str(res))
        return res

    @capcheck(basestring, Decimal, basestring)
    def add_payment(self, payment_method, payment_value, description=''):
        log.info("> CMD: add_payment(method=%r, value=%r, description=%r)" % (payment_method,
                                                                              payment_value,
                                                                              description))

        if not self._has_been_totalized:
            err = "Você deve totalizar o cupom antes de fecha-lo."
            log.error('< ERR: %s' % err)
            raise PaymentAdditionError(_(err))

        res = self._driver.coupon_add_payment(payment_method, payment_value,
                                              self._format_text(description))

        log.info('< RET: %s' % str(res))

        self.payments_total_value += payment_value
        return res

    def open_cash_drawer(self, sec=0):
        log.info('> CMD: open_cash_drawer(sec=%r)' % sec)
        res = self._driver.open_cash_drawer(sec)
        log.info('< RET: %s' % str(res))
        return res

    def get_status_cash_drawer(self):
        log.info('> CMD: get_status_cash_drawer()')
        res = self._driver.get_status_cash_drawer()
        log.info('< RET: %s' % str(res))
        return res

    def cancel(self):
        log.info('> CMD: coupon_cancel()')
        res = self._driver.coupon_cancel()
        log.info('< RET: %s' % str(res))
        self._has_been_totalized = False
        self.payments_total_value = Decimal("0.0")
        self.totalized_value = Decimal("0.0")
        return res

    def cancel_last_coupon(self):
        """Cancel the last non fiscal coupon or the last sale."""
        log.info('> CMD: cancel_last_coupon()')
        res = self._driver.cancel_last_coupon()
        log.info('< RET: %s' % str(res))
        return res

    def _get_last_item_id(self):
        log.info('> CMD: _get_last_item_id()')
        res = self._driver._get_last_item_id()
        log.info('< RET: %s' % str(res))
        return res


    @capcheck(int)
    def cancel_item(self, item_id=None):
        log.info('> CMD: coupon_cancel_item(item_id=%r)' % (item_id,))
        res = self._driver.coupon_cancel_item(item_id)
        log.info('< RET: %s' % str(res))
        return res

    @capcheck(basestring)
    def close(self, promotional_message=''):
        log.info('> CMD: coupon_close(promotional_message=%r)' % promotional_message)

        if not self._has_been_totalized:
            err = "Você deve totalizar o cupom antes de fecha-lo."
            log.error('< ERR: %s' % err)
            raise CloseCouponError(_(err))

        if not self.payments_total_value:
            err = "Não é possível fechar o cupom, pois não existem pagamentos definidos."
            log.error('< ERR: %s' % err)
            raise CloseCouponError(_(err))

        if self.totalized_value > self.payments_total_value:
            err = "Não é possível fechar o cupom porque os pagamentos totais (%.2f) " \
                  "não coicide com o valor totalizado (%.2f)."
            log.error('< ERR: %s' % err)
            raise CloseCouponError(_(err) % (self.payments_total_value, self.totalized_value))

        res = self._driver.coupon_close(self._format_text(promotional_message))
        log.info('< RET: %s' % str(res))

        self.get_status().check_error_two()  # Verifica se teve algum erro na transacao com ECF
        self._has_been_totalized = False
        self.payments_total_value = Decimal("0.0")
        self.totalized_value = Decimal("0.0")

        return res

    def summarize(self):
        log.info('> CMD: summarize()')
        res = self._driver.summarize()
        log.info('< RET: %s' % str(res))
        return res

    def get_totalizadores(self, total=False):
        log.info('> CMD: get_totalizadores(total=%s)' % total)
        res = self._driver.get_totalizadores(total=total)
        log.info('< RET: %s' % str(res))
        return res

    def close_till(self, previous_day=False):
        log.info('> CMD: close_till(previous_day=%r)' % (previous_day,))
        res = self._driver.close_till(previous_day)
        log.info('< RET: %s' % str(res))
        return res

    def _add_voucher(self, type, add_voucher_value, text=''):
        log.info('> CMD: _add_voucher(type=%r, add_voucher_value=%r, text=%r)'
                 % (type, add_voucher_value, text))
        res = self._driver._add_voucher(type, add_voucher_value, text)
        log.info('< RET: %s' % str(res))
        return res

    @capcheck(Decimal, int, int)
    def till_add_cash(self, add_cash_value, totalizador=0, meio=0):
        log.info('> CMD: till_add_cash(add_cash_value=%r, totalizador=%r, meio=%r)'
                 % (add_cash_value, totalizador, meio))
        res = self._driver.till_add_cash(add_cash_value, totalizador=totalizador, meio=meio)
        log.info('< RET: %s' % str(res))
        return res

    @capcheck(Decimal, int, str)
    def till_remove_cash(self, remove_cash_value, cod=0, text=''):
        log.info('> CMD: till_remove_cash(remove_cash_value=%r, cod=%r, text=%r)'
                 % (remove_cash_value, cod, text))

        res = self._driver.till_remove_cash(remove_cash_value, cod=cod, text=text)
        log.info('< RET: %s' % str(res))
        return res

    @capcheck(datetime.date, datetime.date, bool)
    def till_read_memory(self, start, end, full=False):
        assert start <= end <= datetime.date.today(), (
            "start must be less then end and both must be less today")

        log.info('> CMD: till_read_memory(start=%r, end=%r)' % (start, end))

        res = self._driver.till_read_memory(start, end, full)
        log.info('< RET: %s' % str(res))
        return res

    @capcheck(datetime.date, datetime.date, bool)
    def till_read_memory_to_serial(self, start, end, full=False):
        assert start <= end <= datetime.date.today(), (
            "start must be less then end and both must be less today")

        log.info('> CMD: till_read_memory(start=%r, end=%r)' % (start, end))

        res = self._driver.till_read_memory_to_serial(start, end, full)
        log.info('< RET: %s' % str(res))
        return res

    @capcheck(int, int, bool)
    def till_read_memory_by_reductions(self, start, end, full=False):
        assert end >= start > 0, ("start must be less then end and both must be positive")
        log.info('> CMD: till_read_memory_by_reductions(start=%r, end=%r)' % (start, end))

        res = self._driver.till_read_memory_by_reductions(start, end, full)
        log.info('< RET: %s' % str(res))
        return res

    @capcheck(int, int, bool)
    def till_read_memory_by_reductions_to_serial(self, start, end, full=False):
        assert end >= start > 0, ("start must be less then end and both must be positive")
        log.info('> CMD: till_read_memory_by_reductions(start=%r, end=%r)' % (start, end))
        res = self._driver.till_read_memory_by_reductions_to_serial(start, end, full)
        log.info('< RET: %s' % str(res))
        return res

    def gerencial_report_open(self, id=0):
        log.info('> CMD: gerencial_report_open(id=%r)' % id)
        res = self._driver.gerencial_report_open(id=id)
        log.info('< RET: %s' % str(res))
        return res

    def gerencial_report_print(self, text):
        log.info('> CMD: gerencial_report_print(text=%s)' % text)
        res = self._driver.gerencial_report_print(text)
        log.info('< RET: %s' % str(res))
        return res

    def gerencial_report_close(self):
        log.info('> CMD: gerencial_report_open()')
        res = self._driver.gerencial_report_close()
        log.info('< RET: %s' % str(res))
        return res

    def payment_receipt_open(self, identifier, coo, method, value):
        log.info('> CMD: payment_receipt_open(identifier=%s, coo=%s, method=%s, value=%s)'
                 % (identifier, coo, method, value))
        res = self._driver.payment_receipt_open(identifier=identifier, coo=coo, method=method,
                                                value=value)
        log.info('< RET: %s' % str(res))
        return res

    def payment_receipt_print(self, text):
        log.info('> CMD: payment_receipt_print(text=%s)' % text)
        res = self._driver.payment_receipt_print(text)
        log.info('< RET: %s' % str(res))
        return res

    def payment_receipt_close(self):
        log.info('> CMD: payment_receipt_close()')
        res = self._driver.payment_receipt_close()
        log.info('< RET: %s' % str(res))
        return res

    def payment_receipt_print_duplicate(self):
        log.info('> CMD: payment_receipt_print_duplicate()')
        res = self._driver.payment_receipt_print_duplicate()
        log.info('< RET: %s' % str(res))
        return res

    def get_serial(self):
        log.info('> CMD: get_serial()')
        self._serial = self._driver.get_serial() if not self._serial else self._serial
        res = self._serial
        log.info('< RET: %s' % res)
        return res

    def get_serial_mfd(self):
        log.info('> CMD: get_serial()')
        res = self._driver.get_serial_mfd()
        log.info('< RET: %s' % str(res))
        return res

    def query_status(self):
        log.info('> CMD: query_status()')
        res = self._driver.query_status()
        log.info('< RET: %s' % str(res))
        return res

    def send_command(self, command, *args, **kwargs):
        res = self._driver._send_command(command, *args, **kwargs)
        return res

    def get_status(self, val=None, cmd_status=None):
        log.info('> CMD: get_status(val=%r, cmd_status=%r)' % (val, cmd_status))
        res = self._driver.get_status(val, cmd_status)
        log.info('< RET: %s' % str(res))
        return res

    def get_ecf_date(self):
        log.info('> CMD: get_ecf_date()')
        res = self._driver.get_ecf_date()
        log.info('< RET: %s' % str(res))
        return res

    def get_coupon_subtotal(self):
        log.info('> CMD: get_coupon_subtotal()')
        res = self._driver._get_coupon_subtotal()
        log.info('< RET: %s' % str(res))
        return res

    def emite_leitura_fita_detalhe(self, *args, **kwargs):
        log.info('> CMD: emite_leitura_fita_detalhe(%s,%s)' % (args, kwargs))
        res = self._driver._emite_leitura_fita_detalhe(args, kwargs)
        log.info('< RET: %s' % str(res))
        return res

    def le_impressao(self):
        log.info('> CMD: le_impressao()')
        res = self._driver._le_impressao()
        log.info('< RET: %s' % str(res))
        return res

    def set_summertime(self, value=True):
        log.info('> CMD: set_summertime(value=%r)' % value)
        res = self._driver.set_summertime(value)
        log.info('< RET: %s' % str(res))
        return res

    def get_summertime(self):
        log.info('> CMD: get_summertime()')
        res = self._driver.get_summertime()
        log.info('< RET: %s' % str(res))
        return res

    def status_reply_complete(self, reply):
        log.info('> CMD: status_reply_complete(%s)' % (reply,))
        res = self._driver.status_reply_complete(reply)
        log.info('< RET: %s' % str(res))
        return res

    def get_tax_constants(self):
        log.info('> CMD: get_tax_constants()')
        res = self._driver.get_tax_constants()
        log.info('< RET: %s' % str(res))
        return res

    def get_tax(self):
        log.info('> CMD: get_tax()')
        res = self._driver.get_tax()
        log.info('< RET: %s' % res)
        return res

    def get_payment_constants(self):
        log.info('> CMD: get_payment_constants()')
        res = self._driver.get_payment_constants()
        log.info('< RET: %s' % res)
        return res

    def get_payment_receipt_identifier(self, method):
        log.info('> CMD: get_payment_receipt_identifier(method=%s)' % method)
        res = self._driver.get_payment_receipt_identifier(method)
        log.info('< RET: %s' % res)
        return res

    def get_ccf(self):
        """Fiscal Coupon Counter

        @returns: the last document ccf
        @rtype: integer
        """
        log.info('> CMD: get_ccf()')
        res = self._driver.get_ccf()
        log.info('< RET: %s' % res)
        return res

    def get_coo(self):
        """Operation Order Counter

        @returns: the last document coo
        @rtype: integer
        """
        log.info('> CMD: get_coo()')
        res = self._driver.get_coo()
        log.info('< RET: %s' % res)
        return res

    def get_coo_inicial_final(self):
        """Operation Order Counter
            @returns: the last document coo
            @rtype: integer
            """
        log.info('> CMD: get_coo_inicial_final()')
        res = self._driver.get_coo_inicial_final()
        log.info('< RET: %s' % res)
        return res

    def get_read_ie(self):
        """
        @returns: Iscrição estadual
        @return:
        """
        log.info('> CMD: get_read_ie()')
        res = self._driver.get_read_ie()
        log.info('< RET: %s' % res)
        return res

    def get_gnf(self):
        """Nonfiscal Operation General Counter

        @returns: gnf
        @rtype: integer
        """
        log.info('> CMD: get_gnf()')
        res = self._driver.get_gnf()
        log.info('< RET: %s' % res)
        return res

    def get_crz(self):
        """Z Reduction Counter

        @returns: the last document crz
        @rtype: integer
        """
        log.info('> CMD: get_crz()')
        res = self._driver.get_crz()
        log.info('< RET: %s' % res)
        return res

    def get_cfc(self):
        """Cupom Fiscal Cancelado

        @returns: Total de Cupoms Cancelados
        @rtype: integer
        """
        log.info('> CMD: get_cfc()')
        res = self._driver.get_cfc()
        log.info('< RET: %s' % res)
        return res

    def get_cro(self):
        """Cupom de Operacao nao Fiscal

        @returns: Total de operacoes nao fiscais
        @rtype: integer
        """
        log.info('> CMD: get_cro()')
        res = self._driver.get_cro()
        log.info('< RET: %s' % res)
        return res

    def get_grg(self):
        """Contador Geral de Relatório Gerencial

        @returns: Total de Relatório Gerencial
        @rtype: integer
        """
        log.info('> CMD: get_grg()')
        res = self._driver.get_grg()
        log.info('< RET: %s' % res)
        return res

    def get_cdc(self):
        u"""Contador de Comprovantes de Débito ou Crédito

        @returns: Total de Comprovantes de Débito ou Crédito
        @rtype: integer
        """
        log.info('> CMD: get_cdc()')
        res = self._driver.get_cdc()
        log.info('< RET: %s' % res)
        return res

    def get_grande_total(self):
        log.info('> CMD: get_grande_total()')
        res = self._driver.get_grande_total()
        log.info('< RET: %s' % res)
        return res

    def retorno_de_variaveis(self, var, tam):
        log.info('> CMD: retorno_de_variaveis(var%r, tam%r)' % (var, tam))
        res = self._driver.retorno_de_variaveis(var, tam)
        log.info('< RET: %s' % res)
        return res

    def ler_ultima_reducao(self):
        log.info('> CMD: ler_ultima_reducao()')
        res = self._driver.ler_ultima_reducao()
        log.info('< RET: %s' % res)
        return res

    def get_date_last_mov_rz(self):
        log.info('> CMD: get_date_last_mov_rz()')
        res = self._driver.get_date_last_mov_rz()
        log.info('< RET: %s' % res)
        return res

    def get_sintegra(self):
        log.info('> CMD: get_sintegra()')
        res = self._driver.get_sintegra()
        log.info('< RET: %s' % res)
        return res

    def get_data_mov(self):
        log.info('> CMD: get_data_mov()')
        res = self._driver.get_data_mov()
        log.info('< RET: %s' % res)
        return res

    def get_date_sb(self):
        log.info('> CMD: get_date_sb()')
        res = self._driver.get_date_sb()
        log.info('< RET: %s' % res)
        return res

    def get_counters(self):
        log.info('> CMD: get_counters()')
        if self.brand != 'sweda':
            contadores = {
                'coo': self.get_coo(),
                'ccf': self.get_ccf(),
                'gnf': self.get_gnf(),
                'crz': self.get_crz(),
                'grg': self.get_grg(),
                'cdc': self.get_cdc()
            }
        else:
            contadores = self._driver.get_counters()
        log.info('< RET: %s' % contadores)
        return contadores

    def identifier_pafecf(self, text):
        log.info('> CMD: identifier_pafecf()')
        res = self._driver.identifier_pafecf(text)
        log.info('< RET: %s' % res)
        return res

    def get_last_reduce(self):
        log.info('> CMD: get_last_reduce()')
        res = self._driver.get_last_reduce()
        log.info('< RET: %s' % res)
        return res

    def get_num_loja(self):
        log.info('> CMD: get_num_loja()')
        res = self._driver.get_num_loja()
        log.info('< RET: %s' % res)
        return res

    def get_num_ecf(self):
        log.info('> CMD: get_num_ecf()')
        res = self._driver.get_num_ecf()
        log.info('< RET: %s' % res)
        return res

    @property
    def supports_duplicate_receipt(self):
        return self._driver.supports_duplicate_receipt

    @property
    def identify_customer_at_end(self):
        log.info('> CMD: identify_customer_at_end()')
        res = self._driver.identify_customer_at_end
        log.info('< RET: %s' % res)
        return res

    def get_emitir_mfd(self, relatorio, inicio, fim, dir_arq, ecf, tipo='1', dir_bin=''):
        """
        Função especifica para ECF Sweda.
        """
        return self._driver.get_emitir_mfd(relatorio, inicio, fim, dir_arq, ecf, tipo, dir_bin)

    def get_pagamento_existe(self):
        log.info('> CMD: get_pagamento_existe()')
        res = self._driver.get_pagamento_existe()
        log.info('< RET: %s' % res)
        return res