# -*- Mode: Python; coding: utf-8 -*-
# vi:si:et:sw=4:sts=4:ts=4

##
## Stoqdrivers
## Copyright (C) 2005 Async Open Source <http://www.async.com.br>
## All rights reserved
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
## USA.
##
## Author(s):   Evandro Vale Miquelito      <evandro@async.com.br>
##
"""
Sweda IFS9000I driver
"""
import math
import operator
import string
import datetime
import os
from ctypes import CDLL
from decimal import Decimal
import re
from kiwi.python import Settable
from stoqdrivers.logs.log import traced as log
from zope.interface import implements

from stoqdrivers.enum import PaymentMethodType, TaxType, UnitType
from stoqdrivers.exceptions import (PrinterError, CloseCouponError,  PendingReadX, CommandError, CouponOpenError,
                                    CommandParametersError, CouponNotOpenError, ReduceZError, HardwareFailure,
                                    DriverError,  OutofPaperError, PendingReduceZ, InvalidState, CancelItemError,
                                    AlreadyTotalized, PaymentAdditionError, CodigoInvalido )
from stoqdrivers.interfaces import ICouponPrinter
from stoqdrivers.serialbase import SerialBase
from stoqdrivers.printers.capabilities import Capability
from stoqdrivers.printers.base import BaseDriverConstants
from stoqdrivers.translation import stoqdrivers_gettext

_ = stoqdrivers_gettext

root_path = os.path.dirname(os.path.realpath(__file__))

TWOPLACES = Decimal(10) ** -2
TREPLACES = Decimal(10) ** -3

class IFS9000IConstants(BaseDriverConstants):
    _constants = {
        # Page 4-12, two decimals
        UnitType.WEIGHT:      '!',
        UnitType.METERS:      '@',
        UnitType.LITERS:      ')',
        UnitType.EMPTY:       '^',
        PaymentMethodType.MONEY:         '01',
        PaymentMethodType.CHECK:        '01'
        }

    _tax_constants = [
        # Page 4-8
        (TaxType.SUBSTITUTION, 'F', None),
        (TaxType.EXEMPTION,    'I', None),
        (TaxType.NONE,         'N', None),
        # Custom, see Page 4-78
        ]


class IFS9000I(SerialBase):

    implements(ICouponPrinter)

    model_name = "Sweda IFS 9000 I"
    coupon_printer_charset = "ascii"

    CMD_PREFIX = "."
    CMD_SUFFIX = EOL_DELIMIT = '}'

    #
    # Printer command set
    #

    CMD_COUPON_OPEN = '17'
    CMD_COUPON_ADD_ITEM = '01'
    CMD_ITEM_ADD_DISCOUNT = '02'
    CMD_ITEM_ADD_SURCHARGE = '60'
    CMD_ITEM_CANCEL = '04'
    CMD_COUPON_ADD_DISCOUNT = '03'
    CMD_COUPON_ADD_SURCHARGE = '11'
    CMD_COUPON_CANCEL = '05'
    CMD_COUPON_TOTALIZE = '10'
    CMD_COUPON_CLOSE = '12'
    CMD_READ_X = '13'
    CMD_REDUCE_Z = '14'
    CMD_PRINTER_STATUS = '23'
    CMD_PRINT_PARAMETERS = '18'
    CMD_SET_SALE_PARAMETERS = '30'
    CMD_SET_HEADER_PARAMETERS = '31'
    CMD_SAVE_USER_SETTINGS = '41'
    CMD_ADD_USER_SETTINGS = '34'
    CMD_SETUP_CLOCK = '35'
    CMD_SETUP_PAYMENT_METHOD = '39'
    CMD_TRANSACTION_STATUS = '28'
    CMD_LEITURA_DE_TABELAS = '29'
    CMD_GET_PRINTER_TOTALIZERS = '27'
    CMD_OPEN_VOUCHER = '19'
    CMD_INCREMENT_NONFISCAL_COUNTER = '07'
    CMD_IDENTIFIER_PAFECF = '57'
    #
    # Settings for printer command parameters
    #

    CUSTOMER_CHAR_LEN = 20
    CUSTOMER_CNPJ_LEN = 22
    CUSTOMER_STATE_REGISTER_LEN = 21
    # We have actually 24 characters for product descriptions but the first
    # argument define the label for product units
    DESCRIPTION_CHAR_LEN = 23
    # A second description allowed when adding items
    SECONDDESC_CHAR_LEN = 200
    PRODUCT_CODE_CHAR_LEN = 13
    TAXCODE_CHAR_LEN = 3
    PRINT_ONE_LINE_CHAR_LEN = 2
    PERCENTAGE_CHAR_LEN = 4
    ITEM_NUMBER_CHAR_LEN = 3
    PAYMENTMETHOD_CHAR_LEN = 2

    SURCHARGE_CHAR_LEN = 11
    SURCHARGE_DEC_SEPARATOR = 2
    SURCHARGE_ZERO_DIGITS = 0

    PRICE_CHAR_LEN = 8
    PRICE_DEC_SEPARATOR = 2
    PRICE_ZERO_DIGITS = 1

    TOTAL_CHAR_LEN = 8
    TOTAL_DEC_SEPARATOR = 2
    TOTAL_ZERO_DIGITS = 4

    QTY_CHAR_LEN = 7
    QTY_DEC_SEPARATOR= 3
    QTY_ZERO_DIGITS = 0

    # This code must be added in command CMD_COUPON_ADD_SURCHARGE and
    # represents a label 'ACRESCIMO' in the printed coupon
    STANDARD_SURCHARGE_CODE ='51'


    errors_dict = {'DIA ENCERRADO': (PendingReadX, _("Redução X Pendente")),
                   'ERRO-COO NAO EXISTE': (CodigoInvalido, _("Não existe CDC pendente no ECF que "
                                                             "corresponda aos argumentos informados")),
                   'ERRO-COMAND INVALIDO': (CommandError, _("Comando Inválido")),
                   'ERRO-OPERACAO NAO ENCERRADA': (CouponOpenError,
                                                   _("Cupom Aberto")),
                   'ERRO-PARAMETROS DO COMAND INVALIDOS': (CommandParametersError,
                                                           _("Paramêtro Inválido.")),
                   'ERRO-OPERACAO NAO ABERTA': (CouponNotOpenError, 
                                                "Cupom não foi aberto"),
                   'JA  FEZ REDU??O': (ReduceZError, _("Redução Z já efetuada.")),
                   '^RELOGIO ZERADO-PROGRAMAR': (PrinterError,
                                                 _("Printer clock isn't "
                                                   "programmed yet.")),
                   'RELOGIO:ERRO-PROGRAMAR DATA/HORA': (PrinterError,
                                                        _("Printer clock isn't "
                                                          "programmed yet.")),
                   '^CHAMAR ASSISTENCIA TECNICA': (HardwareFailure,
                                                   _("Problem in fiscal printer")),
                   '^CMOS: ERRO-MEM?RIA TAXAS/PARAMET': (HardwareFailure,
                                                         _("Problem in fiscal "
                                                           "printer")),
                   'ERRO-TOTAL: NAO HOUVE LANCAMENTO': (CloseCouponError,
                                                        _("Não há itens adicionado ao cupom")),
                   'ERRO - VERIFICAR PAPEL': (OutofPaperError, _("Out of paper")),
                   'ENCERRAR O DIA !': (PendingReduceZ, _("Redução Z Pendente")),
                   'ERRO-COMANDO NAO PERMITIDO': (InvalidState,
                                                  "Commado não permitido"),
                   'ERRO-SEM OS DADOS DO ITEM NA MEMORIA': (CancelItemError,
                                                            _("It is not "
                                                              "possible cancel "
                                                              "the item since "
                                                              "it wasn't added "
                                                              "yet")),
                   'ERRO-CUPOM TOTALIZADO': (AlreadyTotalized,
                                             _("O Cupom já foi totalizado")),
                   'ERRO-ITEM INEXISTE': (CancelItemError,
                                          _("There is no such item to cancel "
                                            "in the coupon")),
                   'ERRO-CANC: NAO HOUVE LANCAMENTO' : (CouponNotOpenError,
                                                        (_("There's no open document to cancel"))),
                   'ERRO-CODIGO DA MODALIDADE INCORRETO' : (PaymentAdditionError,
                                                            "Finalizadora não cadastrada na ECF."),
                   'ERRO-CODIGO INVALIDO' : (CodigoInvalido, """1. Código do produto igual a zeros.
2. Código do produto igual a espaços não é ISSQN.
3. Código do acumulador não-fiscal não-numérico ou inexistente.
4. Código do acumulador de acréscimo inválido.
5. Código da mensagem informado no comando '<ESC>.09' inexistente.
6. Formato do comando '<ESC>.09' incorreto.""")
    }

    def __init__(self, port, consts=None):
        self._consts = consts or IFS9000IConstants
        SerialBase.__init__(self, port)
        self._customer_name = ''
        self._customer_document = ''
        self._customer_address = ''
        self._customer = ''

    #
    # Initializing Fiscal Printer
    #

    def setup_no_fiscal_legends(self):
        self.send_command("38S&%-14s-%-14s+%-14s", "Caixa", "Sangria",
                          "Suprimento")

    def setup_coupon_header(self, jump_lines_number, header_data):
        """ Call this method only in non-fiscal mode.
        header_data is a list of 5 strings of 40 characters which will be
        added in the header of all the printed coupons.
        """
        command = self.CMD_SET_HEADER_PARAMETERS
        # This is standard. The first argument must be always 'S'
        command += 'S'
        jump_lines_number = str(jump_lines_number)
        if not jump_lines_number.isdigit():
            raise TypeError('Argument jump_lines_number must be numeric')
        if not jump_lines_number.isdigit() in range(10):
            raise TypeError('Argument jump_lines_number must be between '
                            '0 and 9')
        command += jump_lines_number
        # Here we are setting the header lines, five groups where the first
        # element is printer attribute and the second one is a the line
        # format
        if not isinstance(header_data, list):
            raise TypeError('header_data argument must be a list')
        if not len(header_data) == 5:
            raise ValueError('header_data argument must have 5 items')
        for line in header_data:
            # The first element(zero) is a default printer attribute
            command += '0%s' % line
        self.writeline(command)

    def setup_clock(self):
        t = datetime.datetime.now()

        hour = self._format_datetime(t.hour)
        minute = self._format_datetime(t.minute)
        second = self._format_datetime(t.second)
        day = self._format_datetime(t.day)
        month = self._format_datetime(t.month)
        year = str(t.year)[2:]

        time = '%s%s%s' % (hour, minute, second)
        date = '%s%s%s' % (day, month, year)

        command = self.CMD_SETUP_CLOCK + time +  date
        self.writeline(command)

    def setup_sale_parameters(self, use_cents, has_title):
        command = self.CMD_SET_SALE_PARAMETERS
        if use_cents:
            use_cents = 'S'
        else:
            use_cents = 'N'
        if has_title:
            has_title = 'S'
        else:
            has_title = 'N'
        # standard argument for a decimal number.
        # XXX I still don't understand what is it for
        command = (self.CMD_SET_SALE_PARAMETERS + use_cents + '245' +
                   has_title)
        self.writeline(command)

    def setup_printer_user(self, user_cnpj, user_state_reg):
        # XXX Warning: never call this command more than 5 times, otherwise
        # the fiscal print will not work anymore
        command = self.CMD_ADD_USER_SETTINGS
        cnpj = self._format_string(user_cnpj, self.CUSTOMER_CNPJ_LEN,
                                   'user_cnpj')
        state_reg = self._format_string(user_state_reg,
                                        self.CUSTOMER_STATE_REGISTER_LEN,
                                        'user_state_reg')
        command += cnpj + state_reg
        self.writeline(command)

        # Save in memory user information
        command = self.CMD_SAVE_USER_SETTINGS
        # This is a confirm argument. 'S' = confirm, 'N' = cancel
        command += 'S'
        self.writeline(command)

    def setup_sale_data(self):
        """ Call this method only in non-fiscal mode.
        """

        # Set taxes table
        self.writeline('33T010500')

        # Set non-fiscal operation label
        label = '&NAO FISCAL    '
        assert len(label) == 15
        command = '38S%s' % label
        self.writeline(command)

        #
        # Setting up the payment methods
        #
        for method in (PaymentMethodType.MONEY,
                       PaymentMethodType.CHECK):
            label = ("%-15s" % method.get_description())[:15]
            self.send_command(self.CMD_SETUP_PAYMENT_METHOD, 'S', label)

    #
    # Formatting data
    #

    def _check_integer(self, value, arg_name):
        value_str = str(value)
        if not value_str.isdigit():
            raise ValueError('Argument %s must be integer' % arg_name)

    def _format_datetime(self, value):
        value = str(value)
        if len(value) == 1:
            return '0' + value
        return value

    def _format_string(self, value, arg_len, arg_name='', left_justify=False):
        value = str(value)
        if len(value) > arg_len:
            raise ValueError('Argument %s can not have more than %d '
                             'characters' % (arg_name, arg_len))
        if left_justify:
            return string.ljust(value, arg_len)
        return string.rjust(value, arg_len)

    def _format_float(self, value, arg_name, char_len, dec_separator,
                      zero_digits):
        """ - char_len is the maximum amount of character allowed for value
              argument.
            - dec_separator is the number of decimal separator digits
            - zero_digits is the number of zeros we should add in the
              beggining of the argument value.
        """
        value = '%.*f' % (dec_separator, value)

        # As we are going do remove the period character we can here allow
        # char_len + 1 as the maximum number of chars per argument
        value = self._format_string(value, char_len + 1, arg_name)
        value = string.replace(value, ' ', '0')
        value = string.replace(value, '.', '')

        # The first character for total and price arguments must
        # be always zero
        value = '0' * zero_digits + value
        return value

    def _format_price(self, price, arg_name):
        return self._format_float(price, arg_name, self.PRICE_CHAR_LEN,
                                  self.PRICE_DEC_SEPARATOR,
                                  self.PRICE_ZERO_DIGITS)

    def _format_quantity(self, quantity, arg_name):
        return self._format_float(quantity, arg_name, self.QTY_CHAR_LEN,
                                  self.QTY_DEC_SEPARATOR,
                                  self.QTY_ZERO_DIGITS)

    def _format_total(self, total, arg_name):
        return self._format_float(total, arg_name, self.TOTAL_CHAR_LEN,
                                  self.TOTAL_DEC_SEPARATOR,
                                  self.TOTAL_ZERO_DIGITS)

    #
    # Helper methods
    #

    def get_transaction_status(self):
        """28 - Status de Transaçao (pag. 75)"""
        datas = ((3, 4, 'SEQ'), (7, 3, 'ITEM'), (10, 1, 'STATUS'), 
                 (11, 8, 'TRANSACAO'), (19, 2, 'ESCAPE'), (21, 1, 'REDUCAO'),
                 (22, 12, 'LIQ'), (34, 12, 'BRUTO'), (46, 1, 'ERRO'),
                 (47, 6, 'DATA'), (53, 4, 'HORA'), (57, 1, 'VERAO'), 
                 (58, 35, 'MENSERRO'), (93, 3, 'ECF'), (96, 1, 'PAPEL'),
                 (97, 1, 'ABREV'), (98, 12, 'FALTA'), (110, 12, 'REC'),
                 (122, 1, 'FAZX'), (123, 1, 'ABRIR'), (124, 1, 'PROG'),
                 (125, 1, 'VINC'), (126, 1, 'VINC2'), (127, 1, 'SLIP'))
             
        reply = self.writeline(self.CMD_TRANSACTION_STATUS)

        if reply[1] == '+':
            ret_val = {}
            for d in datas:
                idx_ini = d[0] -1 
                idx_end = idx_ini + d[1] 
                data_name = d[2]
                
                ret_val[data_name] = reply[idx_ini:idx_end]
            return ret_val
        return None

    def _reset(self):
        self._customer_name = ''
        self._customer_document = ''
        self._customer_address = ''
        self._customer = ''

    def _get_totalizers(self, type):
        reply = self.send_command(self.CMD_GET_PRINTER_TOTALIZERS, type)
        return reply
    
     
    def get_totalizers(self, type="1"):
        reply = self.writeline(self.CMD_GET_PRINTER_TOTALIZERS + type)
        if reply[1] != '+':
            raise ValueError("Inconsistent package received while "
                             "reading printer totalizers")
        return reply[5:]
    def _send_command(self, command, *args, **kwargs):
        return None
    
    def send_command(self, command, *params):
        """ Send a command to printer.

        command: Is the command in string format
        params: a list of parameter to this command (all parameters
        must be string)
        """
        cmd = command + ''.join(params)

        reply = self.writeline(cmd)

        complete = False
        if command == self.CMD_PRINTER_STATUS:
            return reply
        elif command == '59':
            complete = True
        return self.handle_error(reply, complete=complete)

    def handle_error(self, reply, complete=False):
        # Page 4-2
        # format: |.-NNNNERROR_MESSAGE}|
        if reply[1] != '-' or reply[2] == 'P':
            return reply[2:] if complete else reply
        errmsg = reply[6:]
        try:
            exception, reason = self.errors_dict[errmsg]
            
            raise exception(reason)
        except KeyError:
            raise DriverError(errmsg)

    def get_last_item_id(self):
        # Man page 4-62
        rv = self.get_transaction_status()
        if rv is not None:
            data = rv['ITEM']
            try:
                return int(data)
            except ValueError:
                # This seems to happen the first time of the day we
                # print something.
                log.warn("Could not parse data status item: %r" % data)
        return rv

    def get_totalized_value(self):
        # Man page 4-63
        return self._get_coupon_subtotal()

    def get_remainder_value(self):
        # Man page 4-63
        rv = self.get_transaction_status()
        if rv is not None:
            value = float(rv['BRUTO']) / 1e2
            if value < 0.00:
                return 0.00
        return rv

    def _get_coupon_number(self):
        return int(self.get_totalizers()[115:119])

    def _open_voucher(self, index):
        self.send_command(self.CMD_OPEN_VOUCHER, "%02d" % index)

    def _add_value_to_counter(self, value, counter):

        self.send_command(self.CMD_INCREMENT_NONFISCAL_COUNTER, "%02d" % counter,
                          "%012d" % int(float(value) * 1e2), ' ' * 40)

    #
    # ICouponPrinter implementation
    #

    def coupon_identify_customer(self, customer, address, document):
        # The arguments customer and address are here only for API
        # compatibility, they are not supported by the printer.
        self._customer_name = customer
        self._customer_document = document
        self._customer_address = address

    def coupon_is_customer_identified(self):
        ret = False
        if self._customer_document:
            ret = True 
        return ret 

    def coupon_open(self):
        self.get_status().check_error()
        if self._customer_document:
            self._customer = self._format_string(self._customer_document,
                                           self.CUSTOMER_CHAR_LEN,
                                           'customer_document')
        else:
            self._reset()
        try:
            return self.send_command(self.CMD_COUPON_OPEN, self._customer)
            self._reset()
        except InvalidState:
            # if we catch InvalidState here, probably the printer is with
            # a read X pending, so..
            raise PendingReadX(_("A read X is pending."))


    def arredonda(self, v, d=2):
        ''' O arredondamento segue de acordo com a Norma NBR 5891/77 da ABNT de dezembro de 1977
            está no manual da Sweda
        '''
        parte_1, parte_2 = str(Decimal(v) * Decimal(100.0)).split('.')
        if parte_2[:1] == '5':
            # testa se o proximo algarismo depois do 5 eh 0 ou nao tem nada(que eh zero)
            if len(parte_2) < 2 or parte_2[1:2] == '0':
                # testa se o ultimo numero considerado eh impar, se for soma 1
                if int(parte_1[-1]) % 2 == 1:
                    parte_1 = str(int(parte_1) + 1)
            else:
                parte_1 = str(int(parte_1) + 1)

        elif int(parte_2[:1]) > 5:
            parte_1 = str(int(parte_1) + 1)

        parte_1 = parte_1[:-2] + '.' + parte_1[-2:]
        c = Decimal(10) ** -d
        return Decimal(parte_1).quantize(c)

    def coupon_add_item(self, code, description, price, taxcode,
                        quantity=Decimal("1.0"), unit=UnitType.EMPTY,
                        discount=Decimal("0.0"),
                        surcharge=Decimal("0.0"), unit_desc=""):
        if unit == UnitType.CUSTOM:
            unit = UnitType.EMPTY
        unit_code = self._consts.get_value(unit)
        code = self._format_string(code[-13:], self.PRODUCT_CODE_CHAR_LEN, 'code')
        orig_qty = quantity
        quantity = self._format_quantity(quantity, 'quantity')
        orig_price = price
        price = self._format_price(price, 'price')

        total = self.arredonda(orig_qty * orig_price)

        total = self._format_total(total, 'total')
        if taxcode in ['I', 'F', 'N']:
            taxcode = self._format_string(taxcode, 3, "taxcode", 
                                          left_justify=True)
        else:
            taxcode = 'T%s' % taxcode.rjust(2, '0')
         
        description = str(description)
        if len(description) > self.DESCRIPTION_CHAR_LEN:
            second_desc = description[self.DESCRIPTION_CHAR_LEN:]
            description = description[:self.DESCRIPTION_CHAR_LEN]
        else:
            second_desc = ''
        description = self._format_string(description,
                                          self.DESCRIPTION_CHAR_LEN,
                                          'description')
        if second_desc:
            second_desc = self._format_string(second_desc,
                                              self.SECONDDESC_CHAR_LEN,
                                              'second_desc')

        reply = self.send_command(self.CMD_COUPON_ADD_ITEM, code, quantity,
                                  price, total, unit_code, description, 
                                  taxcode, second_desc)

        if reply[1] == '-' and reply[3:6] == '555':
            raise OutofPaperError('Impressora sem papel!') 
        if discount:
            self.discount_item(discount)
        elif surcharge:
            self.surcharge_item(surcharge)
            
        return self.get_last_item_id()

    def coupon_cancel_item(self, item_id):
        self._check_integer(item_id, 'item_id')
        item_id = self._format_string(item_id, self.ITEM_NUMBER_CHAR_LEN,
                                      'item_id')
        item_id = string.replace(item_id, ' ', '0')

        self.send_command(self.CMD_ITEM_CANCEL, item_id)

    def coupon_add_surcharge(self, item_id, value, description):
        """
        Valid surcharge types are::
          "51": "surcharge"
        The arguments item_id and descriptions only exit for API
        compatibility
        """
        percentage = "0000"
        value *= 10
        value = '%012d' % (Decimal(value) * Decimal('1e2'))

        self.send_command(self.CMD_COUPON_ADD_SURCHARGE, '50', percentage, 
                          value, 'S')

    def coupon_cancel(self):
        self.send_command(self.CMD_COUPON_CANCEL)

    def cancel_last_coupon(self):
        """Cancel the last non fiscal coupon or the last sale."""
        self.send_command(self.CMD_COUPON_CANCEL)

    def coupon_add_payment(self, payment_method, value, description='',
                           custom_pm=''):

        pm = payment_method.rjust(2, '0')
        
        if description:
            description = '{' + description
            
        self.send_command(self.CMD_COUPON_TOTALIZE, pm, '%012d' % int(value * Decimal('1e2')), description)
        return self.get_remainder_value()

    def _configure_payment_methods(self, descricao=u''):
        ind = 0
        finalizadoraNaEcf = {}
        for tipo in ('5', '6', 'Q', 'R'):
            reply = self.send_command(self.CMD_LEITURA_DE_TABELAS, tipo)[7:]
            if  tipo == '5':
                for ret in (reply[29:44], reply[45:60], reply[61:76]):
                    ind += 1
                    finalizadoraNaEcf[ret] = str(ind).rjust(2, '0')
            else:
                i = 16
                x = 1
                while True:
                    print ind
                    if ind == 20:
                        break
                    ret = reply[x:i]
                    x = i+1
                    i += 16
                    ind += 1
                    if len(ret) == 15 and  ret != '               ':
                        finalizadoraNaEcf[ret] = str(ind).rjust(2, '0')
        try:
            result = finalizadoraNaEcf[descricao.ljust(15, ' ')]
        except:
            result =  False
        return result

    def coupon_totalize(self, discount=Decimal("0.0"),
                        surcharge=Decimal("0.0"), taxcode=TaxType.NONE):
        """ Print the total value of the coupon.
        The taxcode argument is useless here and exists only for API
        compatibility
        """

        # FIXME: API changed: discount/surcharge was percentage,
        # now is currency.
        if discount:
            self.discount_coupon(discount)
        elif surcharge:
            self.coupon_add_surcharge(None, surcharge, None)
        self.send_command(self.CMD_COUPON_TOTALIZE)
        return self.get_totalized_value()

    def coupon_close(self, message=''):
        new_message = ''
        if message:
            new_cmd = False
            new_message = ''
            stx_msg = ''
            lines = message.split('\n')
            idx = 0
            for x in range(8):
                if x == 3 and self._customer_name:
                    line = 'CLIENTE: %s' % self._customer_name
                    idx -= 1
                elif x == 4 and self._customer_address:
                    line = 'ENDERECO: %s' % self._customer_address
                    idx -= 1
                else:
                    if idx == len(lines):
                        break
                    else:
                        line = lines[idx]
                        if not line:
                            idx += 1
                            continue
                idx += 1

                if len(line) > 40:
                    new_cmd = True

                stx_msg += line + '\n'
                line = line[:40]

                new_message += '1' + line.ljust(40, ' ')

            if new_message:
                stx_msg = re.sub(r'NOTA LEGAL:', chr(22) + chr(73) + r'NOTA LEGAL:' + chr(22) + chr(65) , stx_msg)
                self.send_command_stx('*07|', stx_msg)
            else:
                self.send_command(self.CMD_COUPON_CLOSE, 'NN', new_message, '}')
            print stx_msg
            print '*' * 60
            print new_message
        self.send_command(self.CMD_COUPON_CLOSE)
        self._customer_document = None
        self._customer_name = ''
        self._customer_address = ''
        return True

    def close_till(self, previous_day=False):
        # TODO Add a date optional argument here
        """This is 'reduce Z' in Brazil"""

        t = datetime.datetime.now()
        day = self._format_datetime(t.day)
        month = self._format_datetime(t.month)
        year = str(t.year)[2:]
        date = '%s%s%s' % (day, month, year)

        # This will print a sales report after summarizing
        print_report = 'N'

        ret_cmd =  self.send_command(self.CMD_REDUCE_Z, print_report, date)
        st_dict = {
            '1': Exception('Impressora off-line'),
            '2': Exception('TIME-OUT de transmissão'),
            '5': Exception('Sem/Papel'),
            }
        if ret_cmd[1] == '-' and ret_cmd[5] in st_dict.keys():
            raise st_dict[ret_cmd[5]]
        return ret_cmd

    def summarize(self):
        """This is 'read X' in Brazil"""
        # This will print a sales report after summarizing
        print_report = 'N'

        self.send_command(self.CMD_READ_X, print_report)

    def till_add_cash(self, value, totalizador, meio):
        # XXX: The manual says that all the printer counters are defined by
        # the user (i.e, can be programmed on the printer memory), so we don't
        # have any counter related to "CASH IN" or "CASH OUT". This part must
        # be considered when fixing bug #2246.  Right now i just consider
        # the counter used by the printer at Async for cash in ("03")
        self._open_voucher(1)
        self._add_value_to_counter(value, totalizador)
        self.coupon_add_payment('%02d' % meio, value)
        self.coupon_close()

    def till_remove_cash(self, value, cod=0, text=''):
        # XXX: The manual says that all the printer counters are defined by
        # the user (i.e, can be programmed on the printer memory), so we don't
        # have any counter related to "CASH IN" or "CASH OUT". This part must
        # be considered when fixing bug #2246.  Right now i just consider
        # the counter used by the printer at Async for cash out ("02")
        self.send_command('19', '%02d' % cod)
        value = '%012d' % (value * Decimal('1e2'))
        self.send_command('07', '%02d' % cod, value)
        self.send_command('12', 'NN2%s' % text)
   
    def calc_checksum(self, data):
        retval = operator.imod(sum([ord(x) for x in data]), 256)
        return chr(retval)
    
    def readline_stx(self):
        # Inicio
        str_ = ''
        buffer = ''
        ler_buffer = 1
        aguardando_checksum = False
        receber_registro = False
        res = ()
        porta = self._port
        while ler_buffer:
            c = porta.readline(1)
            if c: # H� byte lido na porta de comunica��o?
                if aguardando_checksum: # Aguardando o byte de verifica��o (checksum)
                    if self.calc_checksum(buffer) == c: # O byte corresponde ao checksum esperado?
                        porta.write(chr(6)) # Transmitir o caracter ACK (6). Efetivar o registro no buffer
                        res += str_,
                        loop = False
                    else: # Descartar os Dados do registro e transmitir o caracter NAK (21)
                        porta.write(chr(21))
                        buffer = ''
                        str_ = ''
                    aguardando_checksum = False
                else:
                    if c == chr(2): # Trata-se do caracter STX (2)
                        receber_registro = True
                if receber_registro: # H� registro em recep��o?
                    buffer += c
                    if c == chr(0):
                        str_ += ' '
                    else:
                        str_ += c
                    if c == chr(3): # Trata-se do caracter ETX (3)?
                        aguardando_checksum = True
                        receber_registro = False
                else:
                    if c == chr(6):
                        res += str_,
                        ler_buffer = False
                    elif c == chr(21):
                        porta.write(chr(21))
                        aguardando_checksum = False
                        str_ = ''
                        buffer = ''
                        res = ()
                        ler_buffer = False
                    
            else:
                ler_buffer = 0
        return res
    def clear_ecf_buffer(self):
        self._port.flushOutput()
        return 1
        size = 0
        c = '0'
        while c:
            c = self._port.readline(1)
            size += 1
        return size    

    def write_stx(self, data):
        """ Escrever na porta serial """
        print 'Limpar Buffer %d' % self.clear_ecf_buffer() # Limpar o buffer
        self.write(data) # Escreve na porta
        ret_ecf = self._port.readline(1)
        if ret_ecf == chr(6):
            reply = self.readline_stx()
        elif ret_ecf == chr(21):
            reply = ret_ecf
        else:
            print 'Limpar Buffer ERRO %d' % self.clear_ecf_buffer() # Limpar o buffer
            raise Exception("Write data error")
        print 'Limpar Buffer FIM %d' % self.clear_ecf_buffer() # Limpar o buffer
        return reply
    
    def _decompress_data(self, string):
        """ Descomprimi os registros enviados pela impressora 
        
        Formato da compress�o: c ESC n
              onde:
                  c     C�digo asASCII do caracter comprimido
                  ESC   caracter ESC (27), indicador de compress�o
                  n     valor entre 34 e 255 que indica, respectivamente, 3 a 244
                        repeti��es do caracter c
        """
        ret_str = ''
        ignore = False
        for x in range(len(string)):

            c = string[x]
            
            if ignore:
                ignore = False
                print 'ignore', c
                continue
            
            elif ord(c) == 27: 
                n = ord(string[x +1])
                n = n - 34 + 3
                c = string[x -1] * n
                ignore = True
         
                
            ret_str += c
        return ret_str         
    def send_command_stx(self, cmd, *params):
        cmd = chr(2) + cmd + '|'.join(params) + chr(3)
        cmd += self.calc_checksum(cmd)
        
        self.CMD_PREFIX = chr(2)
        self.CMD_SUFFIX = self.EOL_DELIMIT = chr(3)
        print '>>>', cmd
        reply = self.write_stx(cmd)
        
#        print reply
#        read1 = self.read(1)
#        if not read1 == chr(6):
#            raise DriverError('Comando não executado!')
#         
#        reply = self.readline()
        self.CMD_PREFIX = "."
        self.CMD_SUFFIX = self.EOL_DELIMIT = '}'
#        print '>>>', cmd
#        print 'REPLY', reply
#        if reply[4] not in ['+']:
#            print 'err', self.readline()
#            raise DriverError('Comando não executado! ERR[%s]' % reply[5:9])
        return reply
        
        
    def till_read_memory(self, start, end, full=False, serial=False):
        full = 'S' if not full else 'C'
        serial = 'TXT|CP860' if serial else ''
        start = start.strftime('%d%m%Y')
        end = end.strftime('%d%m%Y')
        return self.send_command_stx('*18|', start, end, full, serial)
                
    def till_read_memory_by_reductions(self, start, end, full=False, serial=False):
        full = 'S' if not full else 'C'
        start = '%04d' % start
        end = '%04d' % end
        serial = 'TXT|CP860' if serial else ''
        return self.send_command_stx('*17|', start, end, full, serial)

    def till_read_memory_by_reductions_to_serial(self, start, end, full):   
        reply = self.till_read_memory_by_reductions(start, end, full, True)
        retval = ''
        for x in reply:
            if x[4] == '>':
                retval += unicode('%s\n' % self._decompress_data(x[6:-1]), 'cp860)')
        return retval

    def till_read_memory_to_serial(self, start, end, full):
        reply = self.till_read_memory(start, end, full, True)
        retval = ''
        for x in reply:
            if x[4] == '>':
                retval += unicode('%s\n' % self._decompress_data(x[6:-1]), 'cp860)')
        return retval
    
    def _get_status(self):
        """retornar status de impress�o com string de interpreta��o."""
        return self.send_command(self.CMD_PRINTER_STATUS)

    def get_capabilities(self):
        return dict(item_code=Capability(max_len=14),
                    item_id=Capability(digits=3),
                    items_quantity=Capability(digits=4, decimals=3),
                    item_price=Capability(digits=6, decimals=2),
                    item_description=Capability(max_len=200),
                    payment_value=Capability(digits=10, decimals=2),
                    promotional_message=Capability(max_len=492),
                    payment_description=Capability(max_len=80),
                    customer_name=Capability(max_len=30),
                    customer_id=Capability(max_len=28),
                    customer_address=Capability(max_len=80),
                    add_cash_value=Capability(digits=10, decimals=2),
                    remove_cash_value=Capability(digits=10,
                                                 decimals=2))

    def get_constants(self):
        return self._consts

    #
    # Auxiliar methods
    #

    def discount_coupon(self, value):
        # We are using a only a discount by value for API compatibility
        percentage = '0000'
        value = '%012d' % int(value * Decimal('1e2'))
        # This will print a total right after add the discount
        print_total = 'S'

        self.send_command(self.CMD_COUPON_ADD_DISCOUNT, percentage, value,
                          print_total)

    def desc_acres_item(self, desc_acres, item, valor):
        valor = Decimal(valor)
        if desc_acres == 'D':
            res = self.discount_item(valor, item)
        else:
            res = self.surcharge_item(valor, item)
        return res

    def discount_item(self, value, item=False):
        # We are using a only a discount by value for API compatibility
        percentage = '0000'
        value = '%012d' % (Decimal(value).quantize(TWOPLACES) * Decimal('1e2'))
        parametros = percentage + value
        if item:
            parametros += '%03d' % item
        self.send_command(self.CMD_ITEM_ADD_DISCOUNT, parametros)

    def surcharge_item(self, value, item=False):
        # We are using a only a discount by value for API compatibility
        percentage = '0000' # * self.PERCENTAGE_CHAR_LEN
        value = '%012d' % (value * Decimal('1e2'))
        parametros = percentage + value
        if item:
            parametros += '%03d' % item
        self.send_command(self.CMD_ITEM_ADD_SURCHARGE, parametros)

    def print_printer_parameters(self):
        self.send_command(self.CMD_PRINT_PARAMETERS)

    #### CHRISTIAN #############################################################
    def get_data_movimento(self):
        pass
    def _get_totalizers_g(self):
        return self._get_totalizers_h()
        reply = self._get_totalizers('G')
        print '--', reply

        datas = ((5, 9, 'COO_I'), (9, 13, 'COO'), (13, 17, 'CRO'), 
                 (17, 21, 'CRO_I'), (21, 25, 'CONT_X'))

        if reply:
            ret_val = {}
            for d in datas:
                idx_ini = d[0] 
                idx_end = d[1]  
                data_name = d[2]
                
                ret_val[data_name] = reply[idx_ini:idx_end]
            return ret_val
        return None

    
    def _get_totalizers_h(self):
        reply = self._get_totalizers('H')
        reply = reply[2:]

        datas = ((5, 26, 'SERIAL'), (26, 38, 'SB'), (38, 42, 'CRO'), 
                 (42, 46, 'CRZ'), (46, 52, 'GNF'), (52, 58, 'GRG'), 
                 (58, 64, 'CCF'), (64, 70, 'CFD'), (70, 76, 'COO'),
                 (76, 80, 'CDC'), (80, 84, 'NCN'), (84, 88, 'NFC'),
                 (88, 92, 'CFC'), (92, 110, 'GT'), (110, 124, 'VB'))

        if reply:
            ret_val = {}
            for d in datas:
                idx_ini = d[0] 
                idx_end = d[1]  
                data_name = d[2]
                
                ret_val[data_name] = reply[idx_ini:idx_end]
            return ret_val
        return None
    def _get_coupon_subtotal(self):
        rv = self.get_transaction_status()
        if rv is not None:
            return Decimal(rv['LIQ']) / Decimal("1e2")

    def get_counters(self):
        retval = self._get_totalizers_h()
        ret = {
            'cro': int(retval['CRO']),
            'crz': int(retval['CRZ']),
            'gnf': int(retval['GNF']),
            'grg': int(retval['GRG']),
            'ccf': int(retval['CCF']),
            'cfd': int(retval['CFD']),
            'coo': int(retval['COO']),
            'cdc': int(retval['CDC']),
            'ncn': int(retval['NCN']),
            'nfc': int(retval['NFC']),
            'cfc': int(retval['CFC'])
        }
        return ret
    #****************************
        
    def get_coo(self):
        retval = self._get_totalizers_h()
        return int(retval['COO'])

    def get_cro(self):
        retval = self._get_totalizers_h()
        return int(retval['CRO'])
    
    def get_crz(self):
        retval = self._get_totalizers_h()
        return int(retval['CRZ'])
    
    def get_gnf(self):
        retval = self._get_totalizers_h()
        return int(retval['GNF'])
    
    def get_grg(self):
        retval = self._get_totalizers_h()
        return int(retval['GRG'])
    
    def get_ccf(self):
        retval = self._get_totalizers_h()
        return int(retval['CCF'])
    
    def get_cfd(self):
        retval = self._get_totalizers_h()
        return int(retval['CFD'])
    
    def get_cdc(self):
        retval = self._get_totalizers_h()
        return int(retval['CDC'])
    
    def get_ncn(self):
        retval = self._get_totalizers_h()
        return int(retval['NCN'])
    
    def get_nfc(self):
        retval = self._get_totalizers_h()
        return int(retval['GNF'])

    def get_cfc(self):
        retval = self._get_totalizers_h()
        return int(retval['CFC'])
   
    def get_serial(self):
        retval = self._get_totalizers_h()
        return retval['SERIAL'][:-1]

    
    def get_grande_total(self):
        retval = self._get_totalizers_h()
        return Decimal(int(retval['GT'])) / Decimal("1e2")

    def get_venda_bruta(self):
        retval = self._get_totalizers_h()
        return Decimal(int(retval['VB'])) / Decimal("1e2")

    
    def get_ecf_date(self):
        retval = self.get_transaction_status()
        
        data = retval['DATA']
        horas = retval['HORA']
        
        dia = data[:2]
        mes = data[2:4]
        ano = data [4:]
        hora = horas[:2]
        minutos = horas[2:]
        retval = datetime.datetime(2000 + int(ano), int(mes), int(dia), int(hora), int(minutos), 0)
        return retval

    def get_status(self, val=None, cmd_status=None):
            status = self._get_status()

            sinal = status[1]
            st1 = int(status[3])
            st2 = int(status[4])
            st3 = int(status[5])

            trans = self.get_transaction_status()
            class Status(object):
                def __init__(self, parent=None):
                    self.status = status
                    self.parent = parent
                @property
                def open(self):
                    return trans['STATUS'] == 'P'
                @property
                def pagamento_existe(self):
                    total_pago = ((Decimal(trans['LIQ'])
                                   if trans['LIQ'] else 0 / Decimal("1e2") - Decimal(trans['FALTA'])
                    if trans['FALTA'] else 0 / Decimal("1e2")))
                    return total_pago > 0 and self.open
                @property
                def out_paper(self):
                    return sinal == '-' and st3 == 5
                @property
                def block_z(self):
                    return trans['REDUCAO'] == 'F'
    
                @property
                def off_line(self):
                    return sinal == '-' and st3 == 1

                def check_error(self):
                    if self.block_z:
                        raise PendingReduceZ('Bloqueio por Reduçao Z')
                    if self.open:
                        raise CouponOpenError('Cupom aberto')
                    if st1 and st2 and st3 == 5:
                        raise OutofPaperError('Impressora sem papel.')

                def check_error_two(self):
                    if st1 != 0:
                        self.check_error()

            return Status(self)

    def _apura_totais(self, **kwargs):
        """
        Coment�rios:
        Este comando captura informa��es do equipamento e soma dos
        totais armazenados na Mem�ria Fiscal, conforme per�odo selecionado, 
        aceito nos modelos IF ST200 e IF ST120.
        
        As seguintes informa��es s�o recebidas:
        Modelo, n�mero de fabrica��o, tipo do ECF, n�mero de ordem do ECF,
        n�mero de ordem do usu�rio, CNPJ, Inscri��o Estadual, s�mbolo da
        moeda, n�mero de s�rie da MFD, Grande Total.
        """
        crz_i = crz_f = space = ''
        if kwargs.get('crz_i'):
            if not kwargs.get('crz_f'):
                raise DriverError('Parâmetros do comando inválidos: CRZ final requerido.')
            crz_i = '%04d' % kwargs.get('crz_i')
            crz_f = '%04d' % kwargs.get('crz_f')
            space = '      '

        reply = self.send_command('59', crz_i, space, crz_f, space)
        ret = reply[4:]

        while True:
            reply = self.send_command('++}')
            if reply[:2] == '.]':
                break
            ret += reply[6:]
        try:
            data_mov_rz_str = ret[171:191].split('/') # data do movimento do ultimo Z
            data_fiscal_str = ret[198:208].split('/')
            if data_fiscal_str:
                data_fiscal = datetime.date(int(data_fiscal_str[2]),
                                            int(data_fiscal_str[1]),
                                            int(data_fiscal_str[0]))
            else:
                data_fiscal = False
            if data_mov_rz_str:
                data_mov_z = datetime.date(int(data_fiscal_str[2]),
                                           int(data_fiscal_str[1]),
                                           int(data_fiscal_str[0]))
        except Exception as e:
            raise DriverError('Erro ao apurar totais: %s' % e.message)
        ret_dict = {'DATA_MOV_Z': data_mov_z,
                    'DATA_FISCAL': data_fiscal,
                    'CRZ': int(ret[167:167+4]),
                    'COO_INI': int(ret[209:215]),
                    'CRO_INI': int(ret[215:215+6]),
                    'VBD': Decimal(ret[389:389+18]) / Decimal('100')}
        return ret_dict

    def get_tax(self):
        return self._get_taxes()

    def get_totalizadores(self, total=False):
        retval = []

        ret_cmd = self.send_command('27', '1' )
        if ret_cmd[:3] == '.+C':
            retval.append(
                ('CANC', Decimal(ret_cmd[76:88]) / Decimal('1e2'), 'ICMS')
            )
            retval.append(
                ('DESC', Decimal(ret_cmd[92:104]) / Decimal('1e2'), 'ICMS')
            )


        ret_cmd = self.send_command('27', '3' )
        if ret_cmd[:3] == '.+C':
            retval.append(
                ('I', Decimal(ret_cmd[21:33]) / Decimal('1e2'), 'ICMS')
            )
            retval.append(
                ('N', Decimal(ret_cmd[33:45]) / Decimal('1e2'), 'ICMS')
            )
            retval.append(
                ('F', Decimal(ret_cmd[45:57]) / Decimal('1e2'), 'ICMS')
            )
            it = ret_cmd[93:94] # Tipo Tributa��o, T=ICMS
            ii = ret_cmd[94:96] # Indice da tributa��o
            iv = ret_cmd[96:108] # Valor da tributa��o
            if it == 'T':
                tipo_t = 'ICMS'
            else:
                tipo_t = 'ISS'
            retval.append(
                            (ii, Decimal(iv) / Decimal('1e2'), tipo_t)
                        )
        ret_cmd = ''

        ret_cmd_4 = self.send_command('27', '4' )
        if ret_cmd_4[:3] == '.+C':
            ret_cmd += ret_cmd_4[7:]

        ret_cmd_5 = self.send_command('27', '5' )
        if ret_cmd_5[:3] == '.+C':
            ret_cmd += ret_cmd_5[7:]

        ret_cmd_f = self.send_command('27', 'I' )
        if ret_cmd_f[:3] == '.+C':
            ret_cmd += ret_cmd_f[7:]
        print '*' * 50
        print ret_cmd_f
        print '*' * 50

        if ret_cmd:
            tipo = {}
            i = 0
            for r in range(14):
                reg = ret_cmd[i:i+15]
                i += 15
                if reg[:3] == '   ':
                    continue

                if reg[:1] == 'T':
                    tipo_t = 'ICMS'
                else:
                    tipo_t = 'ISS'

                retval.append((reg[1:3],  Decimal(reg[3:]) / Decimal('1e2'), tipo_t))

        return retval

    def _get_taxes(self):
        
        retval = []
        
        ret_cmd = self.send_command('27', '1' )
        if ret_cmd[:3] == '.+C':
            retval.append(
                ('CANC', Decimal(ret_cmd[76:88]) / Decimal('1e2'), 'ICMS')
            )
            retval.append(
                ('DESC', Decimal(ret_cmd[92:104]) / Decimal('1e2'), 'ICMS')
            )
            
        
        ret_cmd = self.send_command('27', '3' )
        if ret_cmd[:3] == '.+C':
            retval.append(
                ('I', Decimal(ret_cmd[21:33]) / Decimal('1e2'), 'ICMS')
            )
            retval.append(
                ('N', Decimal(ret_cmd[33:45]) / Decimal('1e2'), 'ICMS')
            )
            retval.append(
                ('F', Decimal(ret_cmd[45:57]) / Decimal('1e2'), 'ICMS')
            )
        ret_cmd = ''    

        ret_cmd_4 = self.send_command('27', '4' )
        if ret_cmd_4[:3] == '.+C':
            ret_cmd += ret_cmd_4[7:] 
        
        ret_cmd_5 = self.send_command('27', '5' )
        if ret_cmd_5[:3] == '.+C':
            ret_cmd += ret_cmd_5[7:]

        ret_cmd_f = self.send_command('27', 'I' )
        if ret_cmd_f[:3] == '.+C':
            ret_cmd += ret_cmd_f[7:]
        print '*' * 50
        print ret_cmd_f
        print '*' * 50

        if ret_cmd:
            tipo = {}
            i = 0
            for r in range(14):
                reg = ret_cmd[i:i+15]
                i += 15    
                if reg[:3] == '   ':
                    continue
                try:
                    tipo[reg[:3]] = Decimal(reg[3:]) / Decimal('1e2')
                except:
                    tipo[reg[:3]] = Decimal('000000000000') / Decimal('1e2')
                    pass
            
            ret_cmd_29 = ''            
            ret_cmd_29_3 = self.send_command('29', '3' )
            if ret_cmd_29_3[:3] == '.+T':
                ret_cmd_29 += ret_cmd_29_3[48:48+28]

            ret_cmd_29_4 = self.send_command('29', '4' )
            if ret_cmd_29_4[:3] == '.+T':
                ret_cmd_29 += ret_cmd_29_4[7:]

            ret_cmd_29_5 = self.send_command('29', '5' )
            if ret_cmd_29_5[:3] == '.+T':
                ret_cmd_29 += ret_cmd_29_5[7:28]

            i = 0
            if ret_cmd_29:
                for r in range(14):
                    reg = ret_cmd_29[i:i+7]
                    i += 7
                    it = reg[0] # Tipo Tributa��o, T=ICMS
                    ii = reg[1:3] # Indice da tributa��o
                    iv = reg[3:] # Valor da tributa��o
       
                    if it == ' ':
                        continue
                    getval = tipo.get(it+ii, False)
                    if getval:
                        if it == 'T':
                            tipo_t = 'ICMS'
                        else:
                            tipo_t = 'ISS'
                        retval.append(
                            (iv, getval, tipo_t)
                        )
        return retval
        
    def _get_aliquotas(self, tipo='ICMS'):
        AliquotasStr = '';
        Posicao      = 1;
    
        RetCmd = self.send_command('29', '3' )
        if RetCmd[:3] == '.+T':
            AliquotasStr += '%s|' % RetCmd
        
        RetCmd = self.send_command('29', '4' )
        if RetCmd[:3] == '.+T':
            AliquotasStr += '%s|' % RetCmd

        RetCmd = self.send_command('29', '5' )
        if RetCmd[:3] == '.+T':
            AliquotasStr += RetCmd[7:] 
  
        return AliquotasStr
     
    def get_sintegra(self):
        regs = self._get_totalizers_h()
        if regs:
            data = Settable(
                opening_date=self._apura_totais()['DATA_FISCAL'],
                serial=regs['SERIAL'],
                serial_id=0,
                coupon_start=int(self._apura_totais()['COO_INI']),
                coupon_end=int(regs['COO']),
                cro=int(regs['CRO']),
                crz=int(regs['CRZ']),
                coo=int(regs['COO']),
                ccf=int(regs['CCF']),
                gnf=int(regs['GNF']),
                grg=int(regs['GRG']),
                cdc=int(regs['CDC']),
                period_total=self.get_venda_bruta(),
                total=self.get_grande_total(),
                taxes=self._get_taxes()
            )
            return data
        else:
            return None
    def get_gerencial_reports(self):
        retval = {}
        reply = self.send_command('29', 'M')

        i = 7
        for x in range(8):
            s = reply[i:i+15]
            i += 15
            if s == ' ' * 15:
                continue
            retval[x +1] = s
        return retval
        
    def gerencial_report_open(self, id=0):    
        reports = self.get_gerencial_reports()
        if not reports.get(id, False):
            raise Exception('Relatorio gerencial não cadastrado') 
        self.send_command('13', 'S', reports[id])
        
    def gerencial_report_print(self, text):
        for line in text.split('\n'):
            self.send_command('08', '0', line)

    def gerencial_report_close(self):
        self.send_command('08', '}')

    def open_cash_drawer(self, sec=0):
        return self.send_command('21', '}')

    def get_status_cash_drawer(self):
        """ Status da gaveta:
        .+G001 - Fechada
        .+G000 - Aberta"""
        retcmd = self.send_command('22', '}')
        if not retcmd[1] == '+':
            raise Exception('Não foi possível obter o status da gaveta.')
        return retcmd[5] == '1'

    def payment_receipt_open(self, identifier, coo, method, value):
        #ESC . 1 9 TÍT(2) [COO(4) TIPO(2) IDENT(20) PARC(2)] }
        coo = '%06d' % coo
        self.send_command('19', '00', coo[2:], method)
    
    def payment_receipt_print(self, text):
        # nesta funcao e na funcao 'gerencial_report_print'
        # todo: Verificar o maximo de carcters que pode ser enviado para escrita e pergar o comando para cortar papel
        for line in text.split('\n'):
            self.send_command('08', '0', line)
            
    def payment_receipt_close(self):
        self.send_command('08', '}')

    def set_summertime(self, value):
        p = 'S' if value else 'N'
        return self.send_command('36', p)
        

    def get_summertime(self):
        ret_cmd = self.get_transaction_status()
        ret_val = False if ret_cmd['VERAO'] == 'N' else True
        return ret_val

    def get_data_mov(self):
        retval = self._apura_totais()['DATA_FISCAL']
        return retval

    def get_date_sb(self):
        reply = self.send_command('29', 'H')
        retval = datetime.datetime.strptime(reply[70:80], "%d%m%y%H%M")
        return retval

    def get_date_last_mov_rz(self):
        '''
        :return: Retorna a data do movimento da ultima reducao Z emitida
        '''
        data = 0
        try:
            ultimo_z = self.get_last_reduce()
            if ultimo_z:
                data = ultimo_z.get('DATA_MOV_Z') # data do movimento do ultimo Z emitido

        except Exception as e:
            log.error('get_date_last_mov_rz: %s' % str(e))

        return data

    def get_last_reduce(self):
        crz = int(self._get_totalizers_h()['CRZ']) # Obter o CRZ
        retcmd = self._apura_totais(crz_i=crz, crz_f=crz)
        return {
            'DATA_MOV_Z': retcmd['DATA_MOV_Z'],
            'VBD': retcmd['VBD'],
            'CRO': retcmd['CRO_INI'],
            'CRZ': retcmd['CRZ'],
        }

    def identifier_pafecf(self, text):
        retval = self.send_command(self.CMD_IDENTIFIER_PAFECF, text[:120])
        return retval

    def get_coo_inicial_final(self):
        #Implementar a funcção para sweda
        coo_inicial = '0'
        return coo_inicial
    def get_read_ie(self):
        #Implementar a função para sweda
        return 'isento'

    def get_emitir_mfd(self, relatorio, inicio, fim, dir_arq, ecf, tipo='1', dir_bin=''):
        """
        _______Relatório Espelho MFD_______
        @param tipo: Quando relatório igual Espelho MFD, tipo aceita os valores 0,1,2
        0 = Total
        1 = Por data
        2 =Por COO

        @param inicio: STRING  inicial no formato  data ddmmaa, dd/mm/aa, ddmmaaaa ou dd/mm/aaaa ou o  CRZ/COO inicial.

        @param fim: STRING data fim no formato  ddmmaa, dd/mm/aa, ddmmaaaa ou dd/mm/aaaa ou o  CRZ/COO inicial.

        @param dir_arq: STRING  informando a localização e o nome do arquivo texto a  ser  gerado.

        @param ecf: STRING  informando a código da ecf.

        ________Relatório Arquivo MFD ______
        Tipo:STRING que determina o tipo de relatório a ser gerado:Informar um dos seguintes valores ("0" a "3"):
        "0"=L M F simplificada – formato texto.
        "1"=L M F  COMPLETA    – formato texto.
        "2"=L M F  COMPLETA    – formato COTEPE.
         PatBin = caminho e nome do arquivo de Memória Fiscal no formato binário.

        "2"=M F D – formato COTEPE.Processa o movimento gravado apenas no dispositivo Memória de Fita Detalhe - MFD.
        Para que seja lido apenas a MFD, informar "" ou vazio no argumento abaixo - PatBin.  Assim, NÃO irá processar o movimento do dispositivo da Memória Fiscal.
        A DLL vai procurar o arquivo binário MFD no computador automaticamente.

        "3"=LMF  e  MFD – formato COTEPE.Também conhecido como "TDM" porque inclui movimento da Memória Fiscal e da MFD.
        No argumento abaixo - PatBin -  deverá ser informado o  caminho e o nome do  arquivo  que contém a Memória Fiscal em formato binário.
        Já os arquivos que contém a memória MFD não deverão ser informados. A DLL vai procurar os arquivos binários tipo  MFD  no computador automaticamente.

        "4"=Gera o arquivo com a mesma informação do tipo "3". mas com as seguintes diferenças:
·        A função não consulta o ECF sobre o conteúdo da MFD ·      Não exibe a caixa de geração do arquivo COTEPE. ·
         Poderá usar a pasta informada  no argumento "PatBin" abaixo para obter os arquivos  MFD e gerar o arquivo cotepe na mesma pasta e na pasta informada em "PatTxt".
         Os arquivos não precisam pertencer ao ECF conectado.


        @param inicio: STRING com a data inicial no formato  ddmmaa, dd/mm/aa, ddmmaaaa ou dd/mm/aaaa ou o  CRZ inicial.
        Será assumido que o valor refere-se ao contador COO se o comprimento do valor tiver 7 dígitos.

        @param fim: STRING com a data fim no formato  ddmmaa, dd/mm/aa, ddmmaaaa ou dd/mm/aaaa ou o  CRZ final.
        Será assumido que o valor refere-se ao contador COO se o comprimento do valor tiver 7 dígitos.

        @param dir_arq: STRING  informando a localização e o nome do arquivo texto a  ser  gerado.

        @param dir_bin: STRING  informando a localização e o nome do arquivo binário da Memória Fiscal. Usado para gerar o movimento da Memória Fiscal ou TDM. Este arquivo
        é gerado, chamando a função  ECF_DownloadMF.


        """
        try:
            libSweda =  CDLL('%s/libs/libconvecf.so' %root_path)
            if relatorio == 'Espelho MFD':
                status_mfd = libSweda.ECF_FormatoDadosMFD("",dir_arq, "0", tipo, inicio, fim, ecf)
            elif relatorio == 'Arquivo MFD':
                status_mfd = libSweda.ECF_ReproduzirMemoriaFiscalMFD(tipo,inicio,fim,dir_arq, dir_bin)
            return status_mfd
        except Exception as e:
            print str(e)
            return str(e)
